# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
from unittest import mock

from cryptography import fernet
from oslo_config import cfg
from oslo_config import fixture as oslo_fixture
from oslo_utils import uuidutils
from taskflow.types import failure

from octavia.amphorae.driver_exceptions import exceptions as driver_except
from octavia.common import constants
from octavia.common import data_models
from octavia.common import utils
from octavia.controller.worker.v2.tasks import amphora_driver_tasks
from octavia.db import repositories as repo
from octavia.network import data_models as network_data_models
import octavia.tests.unit.base as base


AMP_ID = uuidutils.generate_uuid()
COMPUTE_ID = uuidutils.generate_uuid()
LISTENER_ID = uuidutils.generate_uuid()
LB_ID = uuidutils.generate_uuid()
CONN_MAX_RETRIES = 10
CONN_RETRY_INTERVAL = 6
FAKE_CONFIG_FILE = 'fake config file'

_db_amphora_mock = mock.MagicMock()
_db_amphora_mock.id = AMP_ID
_db_amphora_mock.status = constants.AMPHORA_ALLOCATED
_db_amphora_mock.vrrp_ip = '198.51.100.65'
_amphora_mock = {
    constants.ID: AMP_ID,
    constants.STATUS: constants.AMPHORA_ALLOCATED,
    constants.COMPUTE_ID: COMPUTE_ID,
}
_db_load_balancer_mock = mock.MagicMock()
_db_load_balancer_mock.id = LB_ID
_listener_mock = mock.MagicMock()
_listener_mock.id = LISTENER_ID
_db_load_balancer_mock.listeners = [_listener_mock]
_vip_mock = mock.MagicMock()
_db_load_balancer_mock.vip = _vip_mock
_LB_mock = {
    constants.LOADBALANCER_ID: LB_ID,
}
_amphorae_mock = [_db_amphora_mock]
_amphora_network_config_mock = mock.MagicMock()
_amphorae_network_config_mock = {
    _amphora_mock[constants.ID]: _amphora_network_config_mock}
_network_mock = mock.MagicMock()
_session_mock = mock.MagicMock()


@mock.patch('octavia.db.repositories.AmphoraRepository.update')
@mock.patch('octavia.db.repositories.AmphoraRepository.get')
@mock.patch('octavia.db.repositories.ListenerRepository.update')
@mock.patch('octavia.db.repositories.ListenerRepository.get',
            return_value=_listener_mock)
@mock.patch('octavia.db.api.get_session', return_value=_session_mock)
@mock.patch('octavia.controller.worker.v2.tasks.amphora_driver_tasks.LOG')
@mock.patch('oslo_utils.uuidutils.generate_uuid', return_value=AMP_ID)
@mock.patch('stevedore.driver.DriverManager.driver')
class TestAmphoraDriverTasks(base.TestCase):

    def setUp(self):

        conf = oslo_fixture.Config(cfg.CONF)
        conf.config(group="haproxy_amphora",
                    active_connection_max_retries=CONN_MAX_RETRIES)
        conf.config(group="haproxy_amphora",
                    active_connection_retry_interval=CONN_RETRY_INTERVAL)
        conf.config(group="controller_worker",
                    loadbalancer_topology=constants.TOPOLOGY_SINGLE)
        self.timeout_dict = {constants.REQ_CONN_TIMEOUT: 1,
                             constants.REQ_READ_TIMEOUT: 2,
                             constants.CONN_MAX_RETRIES: 3,
                             constants.CONN_RETRY_INTERVAL: 4}
        super().setUp()

    @mock.patch('octavia.db.repositories.LoadBalancerRepository.get')
    def test_amp_listeners_update(self,
                                  mock_lb_get,
                                  mock_driver,
                                  mock_generate_uuid,
                                  mock_log,
                                  mock_get_session,
                                  mock_listener_repo_get,
                                  mock_listener_repo_update,
                                  mock_amphora_repo_get,
                                  mock_amphora_repo_update):

        mock_amphora_repo_get.return_value = _db_amphora_mock
        mock_lb_get.return_value = _db_load_balancer_mock
        amp_list_update_obj = amphora_driver_tasks.AmpListenersUpdate()
        amp_list_update_obj.execute(_LB_mock, _amphora_mock, self.timeout_dict)

        mock_driver.update_amphora_listeners.assert_called_once_with(
            _db_load_balancer_mock, _db_amphora_mock, self.timeout_dict)

        mock_driver.update_amphora_listeners.side_effect = Exception('boom')

        amp_list_update_obj.execute(_LB_mock, _amphora_mock, self.timeout_dict)

        mock_amphora_repo_update.assert_called_once_with(
            _session_mock, AMP_ID, status=constants.ERROR)

    @mock.patch('octavia.db.repositories.LoadBalancerRepository.get')
    def test_amp_index_listener_update(self,
                                       mock_lb_get,
                                       mock_driver,
                                       mock_generate_uuid,
                                       mock_log,
                                       mock_get_session,
                                       mock_listener_repo_get,
                                       mock_listener_repo_update,
                                       mock_amphora_repo_get,
                                       mock_amphora_repo_update):

        mock_amphora_repo_get.return_value = _db_amphora_mock
        mock_lb_get.return_value = _db_load_balancer_mock
        amphorae_status = {
            _amphora_mock[constants.ID]: {
                constants.UNREACHABLE: False
            }
        }

        amp_list_update_obj = amphora_driver_tasks.AmphoraIndexListenerUpdate()
        amp_list_update_obj.execute(_LB_mock, 0, [_amphora_mock],
                                    amphorae_status,
                                    _amphora_mock[constants.ID],
                                    self.timeout_dict)

        mock_driver.update_amphora_listeners.assert_called_once_with(
            _db_load_balancer_mock, _db_amphora_mock, self.timeout_dict)

        # Unreachable amp
        mock_driver.reset_mock()
        amphorae_status = {
            _amphora_mock[constants.ID]: {
                constants.UNREACHABLE: True
            }
        }
        amp_list_update_obj.execute(_LB_mock, 0, [_amphora_mock],
                                    amphorae_status,
                                    _amphora_mock[constants.ID],
                                    self.timeout_dict)
        mock_driver.update_amphora_listeners.assert_not_called()

        # Test exception
        mock_driver.update_amphora_listeners.side_effect = Exception('boom')

        amp_list_update_obj.execute(_LB_mock, 0, [_amphora_mock], {},
                                    _amphora_mock[constants.ID],
                                    self.timeout_dict)

        mock_amphora_repo_update.assert_called_once_with(
            _session_mock, AMP_ID, status=constants.ERROR)

        # Test exception, secondary amp
        mock_amphora_repo_update.reset_mock()
        mock_driver.update_amphora_listeners.side_effect = Exception('boom')

        amp_list_update_obj.execute(_LB_mock, 0, [_amphora_mock], {},
                                    '1234',
                                    self.timeout_dict)

        mock_amphora_repo_update.assert_not_called()

    @mock.patch('octavia.db.repositories.LoadBalancerRepository.get')
    @mock.patch('octavia.db.api.session')
    def test_listeners_update(self,
                              mock_get_session_ctx,
                              mock_lb_get,
                              mock_driver,
                              mock_generate_uuid,
                              mock_log,
                              mock_get_session,
                              mock_listener_repo_get,
                              mock_listener_repo_update,
                              mock_amphora_repo_get,
                              mock_amphora_repo_update):
        listeners_update_obj = amphora_driver_tasks.ListenersUpdate()
        LB_ID = 'lb1'
        listeners = [data_models.Listener(id='listener1'),
                     data_models.Listener(id='listener2')]
        vip = data_models.Vip(ip_address='10.0.0.1')
        lb = data_models.LoadBalancer(id=LB_ID, listeners=listeners, vip=vip)
        mock_lb_get.side_effect = [lb, None, lb]
        listeners_update_obj.execute(lb.id)
        mock_driver.update.assert_called_once_with(lb)
        self.assertEqual(1, mock_driver.update.call_count)

        mock_driver.update.reset_mock()

        listeners_update_obj.execute(None)
        mock_driver.update.assert_not_called()

        mock_session = mock_get_session_ctx().begin().__enter__()

        # Test the revert
        amp = listeners_update_obj.revert(_LB_mock)
        expected_db_calls = [mock.call(mock_session,
                                       id=listeners[0].id,
                                       provisioning_status=constants.ERROR),
                             mock.call(mock_session,
                                       id=listeners[1].id,
                                       provisioning_status=constants.ERROR)]
        repo.ListenerRepository.update.assert_has_calls(expected_db_calls)
        self.assertEqual(2, repo.ListenerRepository.update.call_count)
        self.assertIsNone(amp)

    @mock.patch('octavia.db.repositories.LoadBalancerRepository.get')
    @mock.patch('octavia.controller.worker.task_utils.TaskUtils.'
                'mark_listener_prov_status_error')
    def test_amphora_index_listeners_reload(
            self, mock_prov_status_error, mock_lb_repo_get,
            mock_driver, mock_generate_uuid, mock_log, mock_get_session,
            mock_listener_repo_get, mock_listener_repo_update,
            mock_amphora_repo_get, mock_amphora_repo_update):
        listeners_reload_obj = (
            amphora_driver_tasks.AmphoraIndexListenersReload())
        mock_lb = mock.MagicMock()
        mock_listener = mock.MagicMock()
        mock_listener.id = '12345'
        mock_amphora_repo_get.return_value = _amphora_mock
        mock_lb_repo_get.return_value = mock_lb
        mock_driver.reload.side_effect = [mock.DEFAULT, Exception('boom')]

        # Test no listeners
        mock_lb.listeners = None
        listeners_reload_obj.execute(mock_lb, 0, None, {},
                                     _amphora_mock[constants.ID])
        mock_driver.reload.assert_not_called()

        # Test with listeners
        amphorae_status = {
            _amphora_mock[constants.ID]: {
                constants.UNREACHABLE: False
            }
        }
        mock_driver.reload.reset_mock()
        mock_lb.listeners = [mock_listener]
        listeners_reload_obj.execute(mock_lb, 0, [_amphora_mock],
                                     amphorae_status,
                                     _amphora_mock[constants.ID],
                                     timeout_dict=self.timeout_dict)
        mock_driver.reload.assert_called_once_with(mock_lb, _amphora_mock,
                                                   self.timeout_dict)

        # Unreachable amp
        amphorae_status = {
            _amphora_mock[constants.ID]: {
                constants.UNREACHABLE: True
            }
        }
        mock_driver.reload.reset_mock()
        listeners_reload_obj.execute(mock_lb, 0, [_amphora_mock],
                                     amphorae_status,
                                     _amphora_mock[constants.ID],
                                     timeout_dict=self.timeout_dict)
        mock_driver.reload.assert_not_called()

        # Test with reload exception
        mock_driver.reload.reset_mock()
        listeners_reload_obj.execute(mock_lb, 0, [_amphora_mock], {},
                                     _amphora_mock[constants.ID],
                                     timeout_dict=self.timeout_dict)
        mock_driver.reload.assert_called_once_with(mock_lb, _amphora_mock,
                                                   self.timeout_dict)
        mock_amphora_repo_update.assert_called_once_with(
            _session_mock, _amphora_mock[constants.ID],
            status=constants.ERROR)

        # Test with reload exception, secondary amp
        mock_driver.reload.reset_mock()
        mock_amphora_repo_update.reset_mock()
        listeners_reload_obj.execute(mock_lb, 0, [_amphora_mock], {},
                                     '1234',
                                     timeout_dict=self.timeout_dict)
        mock_driver.reload.assert_called_once_with(mock_lb, _amphora_mock,
                                                   self.timeout_dict)
        mock_amphora_repo_update.assert_not_called()

    @mock.patch('octavia.controller.worker.task_utils.TaskUtils.'
                'mark_listener_prov_status_error')
    @mock.patch('octavia.db.repositories.LoadBalancerRepository.get')
    def test_listeners_start(self, mock_lb_get,
                             mock_prov_status_error,
                             mock_driver,
                             mock_generate_uuid,
                             mock_log,
                             mock_get_session,
                             mock_listener_repo_get,
                             mock_listener_repo_update,
                             mock_amphora_repo_get,
                             mock_amphora_repo_update):
        listeners_start_obj = amphora_driver_tasks.ListenersStart()
        mock_lb = mock.MagicMock()
        mock_listener = mock.MagicMock()
        mock_listener.id = '12345'

        # Test no listeners
        mock_lb.listeners = None
        mock_lb_get.return_value = mock_lb
        listeners_start_obj.execute(_LB_mock)
        mock_driver.start.assert_not_called()

        # Test with listeners
        mock_driver.start.reset_mock()
        mock_lb.listeners = [mock_listener]
        listeners_start_obj.execute(_LB_mock)
        mock_driver.start.assert_called_once_with(mock_lb, None)

        # Test revert
        mock_lb.listeners = [mock_listener]
        listeners_start_obj.revert(_LB_mock)
        mock_prov_status_error.assert_called_once_with('12345')

    @mock.patch('octavia.db.api.session')
    def test_listener_delete(self,
                             mock_get_session_ctx,
                             mock_driver,
                             mock_generate_uuid,
                             mock_log,
                             mock_get_session,
                             mock_listener_repo_get,
                             mock_listener_repo_update,
                             mock_amphora_repo_get,
                             mock_amphora_repo_update):

        listener_dict = {constants.LISTENER_ID: LISTENER_ID}
        listener_delete_obj = amphora_driver_tasks.ListenerDelete()
        listener_delete_obj.execute(listener_dict)

        mock_driver.delete.assert_called_once_with(_listener_mock)

        mock_session = mock_get_session_ctx().begin().__enter__()

        # Test the revert
        amp = listener_delete_obj.revert(listener_dict)
        repo.ListenerRepository.update.assert_called_once_with(
            mock_session,
            id=LISTENER_ID,
            provisioning_status=constants.ERROR)
        self.assertIsNone(amp)

        # Test the revert with exception
        repo.ListenerRepository.update.reset_mock()
        mock_listener_repo_update.side_effect = Exception('fail')
        amp = listener_delete_obj.revert(listener_dict)
        repo.ListenerRepository.update.assert_called_once_with(
            mock_session,
            id=LISTENER_ID,
            provisioning_status=constants.ERROR)
        self.assertIsNone(amp)

    def test_amphora_get_info(self,
                              mock_driver,
                              mock_generate_uuid,
                              mock_log,
                              mock_get_session,
                              mock_listener_repo_get,
                              mock_listener_repo_update,
                              mock_amphora_repo_get,
                              mock_amphora_repo_update):

        amphora_get_info_obj = amphora_driver_tasks.AmphoraGetInfo()
        mock_amphora_repo_get.return_value = _db_amphora_mock
        amphora_get_info_obj.execute(_amphora_mock)

        mock_driver.get_info.assert_called_once_with(
            _db_amphora_mock)

    def test_amphora_get_diagnostics(self,
                                     mock_driver,
                                     mock_generate_uuid,
                                     mock_log,
                                     mock_get_session,
                                     mock_listener_repo_get,
                                     mock_listener_repo_update,
                                     mock_amphora_repo_get,
                                     mock_amphora_repo_update):

        amphora_get_diagnostics_obj = (amphora_driver_tasks.
                                       AmphoraGetDiagnostics())
        amphora_get_diagnostics_obj.execute(_amphora_mock)

        mock_driver.get_diagnostics.assert_called_once_with(
            _amphora_mock)

    @mock.patch('octavia.db.api.session')
    def test_amphora_finalize(self,
                              mock_get_session_ctx,
                              mock_driver,
                              mock_generate_uuid,
                              mock_log,
                              mock_get_session,
                              mock_listener_repo_get,
                              mock_listener_repo_update,
                              mock_amphora_repo_get,
                              mock_amphora_repo_update):

        amphora_finalize_obj = amphora_driver_tasks.AmphoraFinalize()
        mock_amphora_repo_get.return_value = _db_amphora_mock
        amphora_finalize_obj.execute(_amphora_mock)

        mock_driver.finalize_amphora.assert_called_once_with(
            _db_amphora_mock)

        mock_session = mock_get_session_ctx().begin().__enter__()

        # Test revert
        amp = amphora_finalize_obj.revert(None, _amphora_mock)
        repo.AmphoraRepository.update.assert_called_once_with(
            mock_session,
            id=AMP_ID,
            status=constants.ERROR)
        self.assertIsNone(amp)

        # Test revert with exception
        repo.AmphoraRepository.update.reset_mock()
        mock_amphora_repo_update.side_effect = Exception('fail')
        amp = amphora_finalize_obj.revert(None, _amphora_mock)
        repo.AmphoraRepository.update.assert_called_once_with(
            mock_session,
            id=AMP_ID,
            status=constants.ERROR)
        self.assertIsNone(amp)

        # Test revert when this task failed
        repo.AmphoraRepository.update.reset_mock()
        amp = amphora_finalize_obj.revert(
            failure.Failure.from_exception(Exception('boom')), _amphora_mock)
        repo.AmphoraRepository.update.assert_not_called()

    @mock.patch('octavia.db.api.session')
    def test_amphora_post_network_plug(self,
                                       mock_get_session_ctx,
                                       mock_driver,
                                       mock_generate_uuid,
                                       mock_log,
                                       mock_get_session,
                                       mock_listener_repo_get,
                                       mock_listener_repo_update,
                                       mock_amphora_repo_get,
                                       mock_amphora_repo_update):

        amphora_post_network_plug_obj = (amphora_driver_tasks.
                                         AmphoraPostNetworkPlug())
        mock_amphora_repo_get.return_value = _db_amphora_mock
        fixed_ips = [{constants.SUBNET: {}}]
        port_mock = {constants.NETWORK: mock.MagicMock(),
                     constants.FIXED_IPS: fixed_ips,
                     constants.ID: uuidutils.generate_uuid()}
        amphora_post_network_plug_obj.execute(_amphora_mock, [port_mock],
                                              _amphora_network_config_mock)

        (mock_driver.post_network_plug.
            assert_called_once_with)(_db_amphora_mock,
                                     network_data_models.Port(**port_mock),
                                     _amphora_network_config_mock)

        mock_session = mock_get_session_ctx().begin().__enter__()

        # Test revert
        amp = amphora_post_network_plug_obj.revert(None, _amphora_mock)
        repo.AmphoraRepository.update.assert_called_once_with(
            mock_session,
            id=AMP_ID,
            status=constants.ERROR)

        self.assertIsNone(amp)

        # Test revert with exception
        repo.AmphoraRepository.update.reset_mock()
        mock_amphora_repo_update.side_effect = Exception('fail')
        amp = amphora_post_network_plug_obj.revert(None, _amphora_mock)
        repo.AmphoraRepository.update.assert_called_once_with(
            mock_session,
            id=AMP_ID,
            status=constants.ERROR)

        self.assertIsNone(amp)

        # Test revert when this task failed
        repo.AmphoraRepository.update.reset_mock()
        amp = amphora_post_network_plug_obj.revert(
            failure.Failure.from_exception(Exception('boom')), _amphora_mock)
        repo.AmphoraRepository.update.assert_not_called()

    def test_amphora_post_network_plug_with_host_routes(
            self, mock_driver, mock_generate_uuid, mock_log, mock_get_session,
            mock_listener_repo_get, mock_listener_repo_update,
            mock_amphora_repo_get, mock_amphora_repo_update):

        amphora_post_network_plug_obj = (amphora_driver_tasks.
                                         AmphoraPostNetworkPlug())
        mock_amphora_repo_get.return_value = _db_amphora_mock
        host_routes = [{'destination': '10.0.0.0/16',
                        'nexthop': '192.168.10.3'},
                       {'destination': '10.2.0.0/16',
                        'nexthop': '192.168.10.5'}]
        fixed_ips = [{constants.SUBNET: {'host_routes': host_routes}}]

        port_mock = {constants.NETWORK: mock.MagicMock(),
                     constants.FIXED_IPS: fixed_ips,
                     constants.ID: uuidutils.generate_uuid()}
        amphora_post_network_plug_obj.execute(_amphora_mock, [port_mock],
                                              _amphora_network_config_mock)

        (mock_driver.post_network_plug.
            assert_called_once_with)(_db_amphora_mock,
                                     network_data_models.Port(**port_mock),
                                     _amphora_network_config_mock)

        call_args = mock_driver.post_network_plug.call_args[0]
        port_arg = call_args[1]
        subnet_arg = port_arg.fixed_ips[0].subnet
        self.assertEqual(2, len(subnet_arg.host_routes))
        for hr1, hr2 in zip(host_routes, subnet_arg.host_routes):
            self.assertEqual(hr1['destination'], hr2.destination)
            self.assertEqual(hr1['nexthop'], hr2.nexthop)

    @mock.patch('octavia.db.repositories.LoadBalancerRepository.get')
    @mock.patch('octavia.db.api.session')
    def test_amphorae_post_network_plug(self,
                                        mock_get_session_ctx,
                                        mock_lb_get,
                                        mock_driver,
                                        mock_generate_uuid,
                                        mock_log,
                                        mock_get_session,
                                        mock_listener_repo_get,
                                        mock_listener_repo_update,
                                        mock_amphora_repo_get,
                                        mock_amphora_repo_update):
        mock_driver.get_network.return_value = _network_mock
        _db_amphora_mock.id = AMP_ID
        _db_amphora_mock.compute_id = COMPUTE_ID
        _db_load_balancer_mock.amphorae = [_db_amphora_mock]
        mock_lb_get.return_value = _db_load_balancer_mock
        mock_amphora_repo_get.return_value = _db_amphora_mock
        amphora_post_network_plug_obj = (amphora_driver_tasks.
                                         AmphoraePostNetworkPlug())

        port_mock = {constants.NETWORK: mock.MagicMock(),
                     constants.FIXED_IPS: [mock.MagicMock()],
                     constants.ID: uuidutils.generate_uuid()}
        _deltas_mock = {_db_amphora_mock.id: [port_mock]}

        amphora_post_network_plug_obj.execute(_LB_mock, _deltas_mock,
                                              _amphorae_network_config_mock)

        (mock_driver.post_network_plug.
         assert_called_once_with(_db_amphora_mock,
                                 network_data_models.Port(**port_mock),
                                 _amphora_network_config_mock))

        # Test with no ports to plug
        mock_driver.post_network_plug.reset_mock()

        _deltas_mock = {'0': [port_mock]}

        amphora_post_network_plug_obj.execute(_LB_mock, _deltas_mock,
                                              _amphora_network_config_mock)
        mock_driver.post_network_plug.assert_not_called()

        mock_session = mock_get_session_ctx().begin().__enter__()

        # Test revert
        amp = amphora_post_network_plug_obj.revert(None, _LB_mock,
                                                   _deltas_mock)
        repo.AmphoraRepository.update.assert_called_once_with(
            mock_session,
            id=AMP_ID,
            status=constants.ERROR)

        self.assertIsNone(amp)

        # Test revert with exception
        repo.AmphoraRepository.update.reset_mock()
        mock_amphora_repo_update.side_effect = Exception('fail')
        amp = amphora_post_network_plug_obj.revert(None, _LB_mock,
                                                   _deltas_mock)
        repo.AmphoraRepository.update.assert_called_once_with(
            mock_session,
            id=AMP_ID,
            status=constants.ERROR)

        self.assertIsNone(amp)

        # Test revert when this task failed
        repo.AmphoraRepository.update.reset_mock()
        amp = amphora_post_network_plug_obj.revert(
            failure.Failure.from_exception(Exception('boom')), _amphora_mock,
            None)
        repo.AmphoraRepository.update.assert_not_called()

    @mock.patch('octavia.db.repositories.LoadBalancerRepository.update')
    @mock.patch('octavia.db.repositories.LoadBalancerRepository.get')
    @mock.patch('octavia.db.api.session')
    def test_amphora_post_vip_plug(self,
                                   mock_get_session_ctx,
                                   mock_lb_get,
                                   mock_loadbalancer_repo_update,
                                   mock_driver,
                                   mock_generate_uuid,
                                   mock_log,
                                   mock_get_session,
                                   mock_listener_repo_get,
                                   mock_listener_repo_update,
                                   mock_amphora_repo_get,
                                   mock_amphora_repo_update):

        amphorae_net_config_mock = {
            AMP_ID: {
                constants.VIP_SUBNET: {
                    'host_routes': []
                },
                constants.VRRP_PORT: mock.MagicMock(),
                'additional_vip_data': []
            }
        }
        mock_amphora_repo_get.return_value = _db_amphora_mock
        mock_lb_get.return_value = _db_load_balancer_mock
        amphora_post_vip_plug_obj = amphora_driver_tasks.AmphoraPostVIPPlug()
        amphora_post_vip_plug_obj.execute(_amphora_mock,
                                          _LB_mock,
                                          amphorae_net_config_mock)
        vip_subnet = network_data_models.Subnet(
            **amphorae_net_config_mock[AMP_ID]['vip_subnet'])
        vrrp_port = network_data_models.Port(
            **amphorae_net_config_mock[AMP_ID]['vrrp_port'])

        mock_driver.post_vip_plug.assert_called_once_with(
            _db_amphora_mock, _db_load_balancer_mock, amphorae_net_config_mock,
            vrrp_port, vip_subnet, additional_vip_data=[])

        mock_session = mock_get_session_ctx().begin().__enter__()

        # Test revert
        amp = amphora_post_vip_plug_obj.revert(None, _amphora_mock, _LB_mock)
        repo.AmphoraRepository.update.assert_called_once_with(
            mock_session,
            id=AMP_ID,
            status=constants.ERROR)
        repo.LoadBalancerRepository.update.assert_not_called()

        self.assertIsNone(amp)

        # Test revert with repo exceptions
        repo.AmphoraRepository.update.reset_mock()
        repo.LoadBalancerRepository.update.reset_mock()
        mock_amphora_repo_update.side_effect = Exception('fail')
        mock_loadbalancer_repo_update.side_effect = Exception('fail')
        amp = amphora_post_vip_plug_obj.revert(None, _amphora_mock, _LB_mock)
        repo.AmphoraRepository.update.assert_called_once_with(
            mock_session,
            id=AMP_ID,
            status=constants.ERROR)
        repo.LoadBalancerRepository.update.assert_not_called()

        self.assertIsNone(amp)

        # Test revert when this task failed
        repo.AmphoraRepository.update.reset_mock()
        amp = amphora_post_vip_plug_obj.revert(
            failure.Failure.from_exception(Exception('boom')), _amphora_mock,
            None)
        repo.AmphoraRepository.update.assert_not_called()

    @mock.patch('octavia.db.repositories.LoadBalancerRepository.update')
    @mock.patch('octavia.db.repositories.LoadBalancerRepository.get')
    def test_amphora_post_vip_plug_with_host_routes(
            self, mock_lb_get, mock_loadbalancer_repo_update, mock_driver,
            mock_generate_uuid, mock_log, mock_get_session,
            mock_listener_repo_get, mock_listener_repo_update,
            mock_amphora_repo_get, mock_amphora_repo_update):

        host_routes = [{'destination': '10.0.0.0/16',
                        'nexthop': '192.168.10.3'},
                       {'destination': '10.2.0.0/16',
                        'nexthop': '192.168.10.5'}]
        amphorae_net_config_mock = {
            AMP_ID: {
                constants.VIP_SUBNET: {
                    'host_routes': host_routes
                },
                constants.VRRP_PORT: mock.MagicMock(),
                'additional_vip_data': []
            }
        }
        mock_amphora_repo_get.return_value = _db_amphora_mock
        mock_lb_get.return_value = _db_load_balancer_mock
        amphora_post_vip_plug_obj = amphora_driver_tasks.AmphoraPostVIPPlug()
        amphora_post_vip_plug_obj.execute(_amphora_mock,
                                          _LB_mock,
                                          amphorae_net_config_mock)
        vip_subnet = network_data_models.Subnet(
            **amphorae_net_config_mock[AMP_ID]['vip_subnet'])
        vrrp_port = network_data_models.Port(
            **amphorae_net_config_mock[AMP_ID]['vrrp_port'])

        mock_driver.post_vip_plug.assert_called_once_with(
            _db_amphora_mock, _db_load_balancer_mock, amphorae_net_config_mock,
            vrrp_port, vip_subnet, additional_vip_data=[])

        call_args = mock_driver.post_vip_plug.call_args[0]
        vip_subnet_arg = call_args[4]
        self.assertEqual(2, len(vip_subnet_arg.host_routes))
        for hr1, hr2 in zip(host_routes, vip_subnet_arg.host_routes):
            self.assertEqual(hr1['destination'], hr2.destination)
            self.assertEqual(hr1['nexthop'], hr2.nexthop)

        self.assertEqual(
            host_routes,
            amphorae_net_config_mock[AMP_ID][
                constants.VIP_SUBNET]['host_routes'])

    @mock.patch('octavia.db.repositories.LoadBalancerRepository.update')
    @mock.patch('octavia.db.repositories.LoadBalancerRepository.get')
    def test_amphora_post_vip_plug_with_additional_vips(
            self, mock_lb_get, mock_loadbalancer_repo_update, mock_driver,
            mock_generate_uuid, mock_log, mock_get_session,
            mock_listener_repo_get, mock_listener_repo_update,
            mock_amphora_repo_get, mock_amphora_repo_update):

        host_routes = [{'destination': '10.0.0.0/16',
                        'nexthop': '192.168.10.3'},
                       {'destination': '10.2.0.0/16',
                        'nexthop': '192.168.10.5'}]
        additional_host_routes = [{'destination': '2001:db9::/64',
                                   'nexthop': '2001:db8::1:fff'}]
        amphorae_net_config_mock = {
            AMP_ID: {
                constants.VIP_SUBNET: {
                    'host_routes': host_routes
                },
                constants.VRRP_PORT: mock.MagicMock(),
                'additional_vip_data': [{
                    'ip_address': '2001:db8::3',
                    'subnet': {
                        'host_routes': additional_host_routes
                    }
                }]
            }
        }
        mock_amphora_repo_get.return_value = _db_amphora_mock
        mock_lb_get.return_value = _db_load_balancer_mock
        amphora_post_vip_plug_obj = amphora_driver_tasks.AmphoraPostVIPPlug()
        amphora_post_vip_plug_obj.execute(_amphora_mock,
                                          _LB_mock,
                                          amphorae_net_config_mock)
        vip_subnet = network_data_models.Subnet(
            **amphorae_net_config_mock[AMP_ID]['vip_subnet'])
        vrrp_port = network_data_models.Port(
            **amphorae_net_config_mock[AMP_ID]['vrrp_port'])
        additional_vip_data = [
            network_data_models.AdditionalVipData(
                ip_address=add_vip_data['ip_address'],
                subnet=network_data_models.Subnet(
                    host_routes=add_vip_data['subnet']['host_routes']))
            for add_vip_data in amphorae_net_config_mock[
                AMP_ID]['additional_vip_data']]

        mock_driver.post_vip_plug.assert_called_once_with(
            _db_amphora_mock, _db_load_balancer_mock, amphorae_net_config_mock,
            vrrp_port, vip_subnet, additional_vip_data=additional_vip_data)

        call_args = mock_driver.post_vip_plug.call_args[0]
        call_kwargs = mock_driver.post_vip_plug.call_args[1]
        vip_subnet_arg = call_args[4]
        self.assertEqual(2, len(vip_subnet_arg.host_routes))
        for hr1, hr2 in zip(host_routes, vip_subnet_arg.host_routes):
            self.assertEqual(hr1['destination'], hr2.destination)
            self.assertEqual(hr1['nexthop'], hr2.nexthop)

        self.assertEqual(
            host_routes,
            amphorae_net_config_mock[AMP_ID][
                constants.VIP_SUBNET]['host_routes'])

        add_vip_data_arg = call_kwargs.get('additional_vip_data')
        self.assertEqual(1, len(add_vip_data_arg[0].subnet.host_routes))
        hr1 = add_vip_data_arg[0].subnet.host_routes[0]
        self.assertEqual(
            additional_host_routes[0]['destination'], hr1.destination)
        self.assertEqual(
            additional_host_routes[0]['nexthop'], hr1.nexthop)

    @mock.patch('octavia.db.repositories.LoadBalancerRepository.update')
    @mock.patch('octavia.db.repositories.LoadBalancerRepository.get')
    def test_amphorae_post_vip_plug(self, mock_lb_get,
                                    mock_loadbalancer_repo_update,
                                    mock_driver,
                                    mock_generate_uuid,
                                    mock_log,
                                    mock_get_session,
                                    mock_listener_repo_get,
                                    mock_listener_repo_update,
                                    mock_amphora_repo_get,
                                    mock_amphora_repo_update):

        amphorae_net_config_mock = mock.MagicMock()
        mock_amphora_repo_get.return_value = _db_amphora_mock
        vip_subnet = network_data_models.Subnet(
            **amphorae_net_config_mock[AMP_ID]['vip_subnet'])
        vrrp_port = network_data_models.Port(
            **amphorae_net_config_mock[AMP_ID]['vrrp_port'])
        _db_load_balancer_mock.amphorae = [_db_amphora_mock]
        mock_lb_get.return_value = _db_load_balancer_mock
        amphora_post_vip_plug_obj = amphora_driver_tasks.AmphoraePostVIPPlug()
        amphora_post_vip_plug_obj.execute(_LB_mock,
                                          amphorae_net_config_mock)

        mock_driver.post_vip_plug.assert_called_once_with(
            _db_amphora_mock, _db_load_balancer_mock, amphorae_net_config_mock,
            vrrp_port, vip_subnet, additional_vip_data=[])

    def test_amphora_cert_upload(self,
                                 mock_driver,
                                 mock_generate_uuid,
                                 mock_log,
                                 mock_get_session,
                                 mock_listener_repo_get,
                                 mock_listener_repo_update,
                                 mock_amphora_repo_get,
                                 mock_amphora_repo_update):
        key = utils.get_compatible_server_certs_key_passphrase()
        mock_amphora_repo_get.return_value = _db_amphora_mock
        fer = fernet.Fernet(key)
        pem_file_mock = fer.encrypt(
            utils.get_compatible_value('test-pem-file')).decode('utf-8')
        amphora_cert_upload_mock = amphora_driver_tasks.AmphoraCertUpload()
        amphora_cert_upload_mock.execute(_amphora_mock, pem_file_mock)

        mock_driver.upload_cert_amp.assert_called_once_with(
            _db_amphora_mock, fer.decrypt(pem_file_mock.encode('utf-8')))

    def test_amphora_update_vrrp_interface(self,
                                           mock_driver,
                                           mock_generate_uuid,
                                           mock_log,
                                           mock_get_session,
                                           mock_listener_repo_get,
                                           mock_listener_repo_update,
                                           mock_amphora_repo_get,
                                           mock_amphora_repo_update):
        FAKE_INTERFACE = 'fake0'
        mock_amphora_repo_get.return_value = _db_amphora_mock
        mock_driver.get_interface_from_ip.side_effect = [FAKE_INTERFACE,
                                                         Exception('boom')]

        timeout_dict = {constants.CONN_MAX_RETRIES: CONN_MAX_RETRIES,
                        constants.CONN_RETRY_INTERVAL: CONN_RETRY_INTERVAL}

        amphora_update_vrrp_interface_obj = (
            amphora_driver_tasks.AmphoraUpdateVRRPInterface())
        amphora_update_vrrp_interface_obj.execute(_amphora_mock, timeout_dict)
        mock_driver.get_interface_from_ip.assert_called_once_with(
            _db_amphora_mock, _db_amphora_mock.vrrp_ip,
            timeout_dict=timeout_dict)
        mock_amphora_repo_update.assert_called_once_with(
            _session_mock, _db_amphora_mock.id, vrrp_interface=FAKE_INTERFACE)

        # Test with an exception
        mock_amphora_repo_update.reset_mock()
        amphora_update_vrrp_interface_obj.execute(_amphora_mock, timeout_dict)
        mock_amphora_repo_update.assert_called_once_with(
            _session_mock, _db_amphora_mock.id, status=constants.ERROR)

    def test_amphora_index_update_vrrp_interface(
            self, mock_driver, mock_generate_uuid, mock_log, mock_get_session,
            mock_listener_repo_get, mock_listener_repo_update,
            mock_amphora_repo_get, mock_amphora_repo_update):
        mock_amphora_repo_get.return_value = _db_amphora_mock
        FAKE_INTERFACE = 'fake0'
        mock_driver.get_interface_from_ip.side_effect = [FAKE_INTERFACE,
                                                         Exception('boom')]
        amphorae_status = {
            _amphora_mock[constants.ID]: {
                constants.UNREACHABLE: False
            }
        }

        timeout_dict = {constants.CONN_MAX_RETRIES: CONN_MAX_RETRIES,
                        constants.CONN_RETRY_INTERVAL: CONN_RETRY_INTERVAL}

        amphora_update_vrrp_interface_obj = (
            amphora_driver_tasks.AmphoraIndexUpdateVRRPInterface())
        amphora_update_vrrp_interface_obj.execute(
            0, [_amphora_mock], amphorae_status, _amphora_mock[constants.ID],
            timeout_dict)
        mock_driver.get_interface_from_ip.assert_called_once_with(
            _db_amphora_mock, _db_amphora_mock.vrrp_ip,
            timeout_dict=timeout_dict)
        mock_amphora_repo_update.assert_called_once_with(
            _session_mock, _db_amphora_mock.id, vrrp_interface=FAKE_INTERFACE)

        # Unreachable amp
        mock_driver.reset_mock()
        amphorae_status = {
            _amphora_mock[constants.ID]: {
                constants.UNREACHABLE: True
            }
        }
        amphora_update_vrrp_interface_obj.execute(
            0, [_amphora_mock], amphorae_status, _amphora_mock[constants.ID],
            timeout_dict)
        mock_driver.get_interface_from_ip.assert_not_called()

        # Test with an exception
        mock_amphora_repo_update.reset_mock()
        amphora_update_vrrp_interface_obj.execute(
            0, [_amphora_mock], {}, _amphora_mock[constants.ID], timeout_dict)
        mock_amphora_repo_update.assert_called_once_with(
            _session_mock, _db_amphora_mock.id, status=constants.ERROR)

        # Test with an exception, secondary amp
        mock_amphora_repo_update.reset_mock()
        amphora_update_vrrp_interface_obj.execute(
            0, [_amphora_mock], {}, '1234', timeout_dict)
        mock_amphora_repo_update.assert_not_called()

    @mock.patch('octavia.db.repositories.LoadBalancerRepository.get')
    def test_amphora_vrrp_update(self,
                                 mock_lb_get,
                                 mock_driver,
                                 mock_generate_uuid,
                                 mock_log,
                                 mock_get_session,
                                 mock_listener_repo_get,
                                 mock_listener_repo_update,
                                 mock_amphora_repo_get,
                                 mock_amphora_repo_update):
        amphorae_network_config = mock.MagicMock()
        mock_driver.update_vrrp_conf.side_effect = [mock.DEFAULT,
                                                    Exception('boom')]
        mock_lb_get.return_value = _db_load_balancer_mock
        mock_amphora_repo_get.return_value = _db_amphora_mock
        amphora_vrrp_update_obj = (
            amphora_driver_tasks.AmphoraVRRPUpdate())
        amphora_vrrp_update_obj.execute(LB_ID, amphorae_network_config,
                                        _amphora_mock, 'fakeint0')
        mock_driver.update_vrrp_conf.assert_called_once_with(
            _db_load_balancer_mock, amphorae_network_config,
            _db_amphora_mock, None)

        # Test with an exception
        mock_amphora_repo_update.reset_mock()
        amphora_vrrp_update_obj.execute(LB_ID, amphorae_network_config,
                                        _amphora_mock, 'fakeint0')
        mock_amphora_repo_update.assert_called_once_with(
            _session_mock, _db_amphora_mock.id, status=constants.ERROR)

    @mock.patch('octavia.db.repositories.LoadBalancerRepository.get')
    def test_amphora_index_vrrp_update(self,
                                       mock_lb_get,
                                       mock_driver,
                                       mock_generate_uuid,
                                       mock_log,
                                       mock_get_session,
                                       mock_listener_repo_get,
                                       mock_listener_repo_update,
                                       mock_amphora_repo_get,
                                       mock_amphora_repo_update):
        amphorae_network_config = mock.MagicMock()
        mock_driver.update_vrrp_conf.side_effect = [mock.DEFAULT,
                                                    Exception('boom')]
        mock_lb_get.return_value = _db_load_balancer_mock
        mock_amphora_repo_get.return_value = _db_amphora_mock
        amphorae_status = {
            _amphora_mock[constants.ID]: {
                constants.UNREACHABLE: False
            }
        }

        amphora_vrrp_update_obj = (
            amphora_driver_tasks.AmphoraIndexVRRPUpdate())

        amphora_vrrp_update_obj.execute(LB_ID, amphorae_network_config,
                                        0, [_amphora_mock], amphorae_status,
                                        'fakeint0',
                                        _amphora_mock[constants.ID],
                                        timeout_dict=self.timeout_dict)
        mock_driver.update_vrrp_conf.assert_called_once_with(
            _db_load_balancer_mock, amphorae_network_config, _db_amphora_mock,
            self.timeout_dict)

        # Unreachable amp
        amphorae_status = {
            _amphora_mock[constants.ID]: {
                constants.UNREACHABLE: True
            }
        }
        mock_amphora_repo_update.reset_mock()
        mock_driver.update_vrrp_conf.reset_mock()
        amphora_vrrp_update_obj.execute(LB_ID, amphorae_network_config,
                                        0, [_amphora_mock], amphorae_status,
                                        None, _amphora_mock[constants.ID])
        mock_driver.update_vrrp_conf.assert_not_called()

        # Test with an exception
        mock_amphora_repo_update.reset_mock()
        amphora_vrrp_update_obj.execute(LB_ID, amphorae_network_config,
                                        0, [_amphora_mock], {}, 'fakeint0',
                                        _amphora_mock[constants.ID])
        mock_amphora_repo_update.assert_called_once_with(
            _session_mock, _db_amphora_mock.id, status=constants.ERROR)

        # Test with an exception, secondary amp
        mock_amphora_repo_update.reset_mock()
        amphora_vrrp_update_obj.execute(LB_ID, amphorae_network_config,
                                        0, [_amphora_mock], {}, 'fakeint0',
                                        '1234')
        mock_amphora_repo_update.assert_not_called()

    def test_amphora_vrrp_start(self,
                                mock_driver,
                                mock_generate_uuid,
                                mock_log,
                                mock_get_session,
                                mock_listener_repo_get,
                                mock_listener_repo_update,
                                mock_amphora_repo_get,
                                mock_amphora_repo_update):
        mock_amphora_repo_get.return_value = _db_amphora_mock
        amphora_vrrp_start_obj = (
            amphora_driver_tasks.AmphoraVRRPStart())
        amphora_vrrp_start_obj.execute(_amphora_mock,
                                       timeout_dict=self.timeout_dict)
        mock_driver.start_vrrp_service.assert_called_once_with(
            _db_amphora_mock, self.timeout_dict)

    def test_amphora_index_vrrp_start(self,
                                      mock_driver,
                                      mock_generate_uuid,
                                      mock_log,
                                      mock_get_session,
                                      mock_listener_repo_get,
                                      mock_listener_repo_update,
                                      mock_amphora_repo_get,
                                      mock_amphora_repo_update):
        mock_amphora_repo_get.return_value = _db_amphora_mock
        amphorae_status = {
            _amphora_mock[constants.ID]: {
                constants.UNREACHABLE: False
            }
        }

        amphora_vrrp_start_obj = (
            amphora_driver_tasks.AmphoraIndexVRRPStart())
        mock_driver.start_vrrp_service.side_effect = [mock.DEFAULT,
                                                      Exception('boom')]

        amphora_vrrp_start_obj.execute(0, [_amphora_mock], amphorae_status,
                                       _amphora_mock[constants.ID],
                                       timeout_dict=self.timeout_dict)
        mock_driver.start_vrrp_service.assert_called_once_with(
            _db_amphora_mock, self.timeout_dict)

        # Unreachable amp
        mock_driver.start_vrrp_service.reset_mock()
        amphorae_status = {
            _amphora_mock[constants.ID]: {
                constants.UNREACHABLE: True
            }
        }
        amphora_vrrp_start_obj.execute(0, [_amphora_mock], amphorae_status,
                                       _amphora_mock[constants.ID],
                                       timeout_dict=self.timeout_dict)
        mock_driver.start_vrrp_service.assert_not_called()

        # Test with a start exception
        mock_driver.start_vrrp_service.reset_mock()
        amphora_vrrp_start_obj.execute(0, [_amphora_mock], {},
                                       _amphora_mock[constants.ID],
                                       timeout_dict=self.timeout_dict)
        mock_driver.start_vrrp_service.assert_called_once_with(
            _db_amphora_mock, self.timeout_dict)
        mock_amphora_repo_update.assert_called_once_with(
            _session_mock, _db_amphora_mock.id, status=constants.ERROR)

        # Test with a start exception, secondary amp
        mock_driver.start_vrrp_service.reset_mock()
        mock_amphora_repo_update.reset_mock()
        amphora_vrrp_start_obj.execute(0, [_amphora_mock], {}, '1234',
                                       timeout_dict=self.timeout_dict)
        mock_driver.start_vrrp_service.assert_called_once_with(
            _db_amphora_mock, self.timeout_dict)
        mock_amphora_repo_update.assert_not_called()

    def test_amphora_compute_connectivity_wait(self,
                                               mock_driver,
                                               mock_generate_uuid,
                                               mock_log,
                                               mock_get_session,
                                               mock_listener_repo_get,
                                               mock_listener_repo_update,
                                               mock_amphora_repo_get,
                                               mock_amphora_repo_update):
        amp_compute_conn_wait_obj = (
            amphora_driver_tasks.AmphoraComputeConnectivityWait())
        mock_amphora_repo_get.return_value = _db_amphora_mock
        amp_compute_conn_wait_obj.execute(_amphora_mock,
                                          raise_retry_exception=True)
        mock_driver.get_info.assert_called_once_with(
            _db_amphora_mock, raise_retry_exception=True)

        mock_driver.get_info.side_effect = driver_except.TimeOutException()
        self.assertRaises(driver_except.TimeOutException,
                          amp_compute_conn_wait_obj.execute, _amphora_mock)
        mock_amphora_repo_update.assert_called_once_with(
            _session_mock, AMP_ID, status=constants.ERROR)

    @mock.patch('octavia.amphorae.backends.agent.agent_jinja_cfg.'
                'AgentJinjaTemplater.build_agent_config')
    def test_amphora_config_update(self,
                                   mock_build_config,
                                   mock_driver,
                                   mock_generate_uuid,
                                   mock_log,
                                   mock_get_session,
                                   mock_listener_repo_get,
                                   mock_listener_repo_update,
                                   mock_amphora_repo_get,
                                   mock_amphora_repo_update):
        mock_build_config.return_value = FAKE_CONFIG_FILE
        amp_config_update_obj = amphora_driver_tasks.AmphoraConfigUpdate()
        mock_driver.update_amphora_agent_config.side_effect = [
            None, None, driver_except.AmpDriverNotImplementedError,
            driver_except.TimeOutException]
        # With Flavor
        flavor = {constants.LOADBALANCER_TOPOLOGY:
                  constants.TOPOLOGY_ACTIVE_STANDBY}
        mock_amphora_repo_get.return_value = _db_amphora_mock
        amp_config_update_obj.execute(_amphora_mock, flavor)
        mock_build_config.assert_called_once_with(
            _db_amphora_mock.id, constants.TOPOLOGY_ACTIVE_STANDBY)
        mock_driver.update_amphora_agent_config.assert_called_once_with(
            _db_amphora_mock, FAKE_CONFIG_FILE)
        # With no Flavor
        mock_driver.reset_mock()
        mock_build_config.reset_mock()
        amp_config_update_obj.execute(_amphora_mock, None)
        mock_build_config.assert_called_once_with(
            _db_amphora_mock.id, constants.TOPOLOGY_SINGLE)
        mock_driver.update_amphora_agent_config.assert_called_once_with(
            _db_amphora_mock, FAKE_CONFIG_FILE)
        # With amphora that does not support config update
        mock_driver.reset_mock()
        mock_build_config.reset_mock()
        amp_config_update_obj.execute(_amphora_mock, flavor)
        mock_build_config.assert_called_once_with(
            _db_amphora_mock.id, constants.TOPOLOGY_ACTIVE_STANDBY)
        mock_driver.update_amphora_agent_config.assert_called_once_with(
            _db_amphora_mock, FAKE_CONFIG_FILE)
        # With an unknown exception
        mock_driver.reset_mock()
        mock_build_config.reset_mock()
        self.assertRaises(driver_except.TimeOutException,
                          amp_config_update_obj.execute,
                          _amphora_mock, flavor)

    def test_amphorae_get_connectivity_status(self,
                                              mock_driver,
                                              mock_generate_uuid,
                                              mock_log,
                                              mock_get_session,
                                              mock_listener_repo_get,
                                              mock_listener_repo_update,
                                              mock_amphora_repo_get,
                                              mock_amphora_repo_update):
        amphora1_mock = mock.MagicMock()
        amphora1_mock[constants.ID] = 'id1'
        amphora2_mock = mock.MagicMock()
        amphora2_mock[constants.ID] = 'id2'
        db_amphora1_mock = mock.Mock()
        db_amphora2_mock = mock.Mock()

        amp_get_connectivity_status = (
            amphora_driver_tasks.AmphoraeGetConnectivityStatus())

        # All amphorae reachable
        mock_amphora_repo_get.side_effect = [
            db_amphora1_mock,
            db_amphora2_mock]
        mock_driver.check.return_value = None

        ret = amp_get_connectivity_status.execute(
            [amphora1_mock, amphora2_mock],
            amphora1_mock[constants.ID],
            timeout_dict=self.timeout_dict)
        mock_driver.check.assert_has_calls(
            [mock.call(db_amphora1_mock, timeout_dict=self.timeout_dict),
             mock.call(db_amphora2_mock, timeout_dict=self.timeout_dict)])
        self.assertFalse(
            ret[amphora1_mock[constants.ID]][constants.UNREACHABLE])
        self.assertFalse(
            ret[amphora2_mock[constants.ID]][constants.UNREACHABLE])

        # amphora1 unreachable
        mock_driver.check.reset_mock()
        mock_amphora_repo_get.side_effect = [
            db_amphora1_mock,
            db_amphora2_mock]
        mock_driver.check.side_effect = [
            driver_except.TimeOutException, None]
        self.assertRaises(driver_except.TimeOutException,
                          amp_get_connectivity_status.execute,
                          [amphora1_mock, amphora2_mock],
                          amphora1_mock[constants.ID],
                          timeout_dict=self.timeout_dict)
        mock_driver.check.assert_called_with(
            db_amphora1_mock, timeout_dict=self.timeout_dict)

        # amphora2 unreachable
        mock_driver.check.reset_mock()
        mock_amphora_repo_get.side_effect = [
            db_amphora1_mock,
            db_amphora2_mock]
        mock_driver.check.side_effect = [
            None, driver_except.TimeOutException]
        ret = amp_get_connectivity_status.execute(
            [amphora1_mock, amphora2_mock],
            amphora1_mock[constants.ID],
            timeout_dict=self.timeout_dict)
        mock_driver.check.assert_has_calls(
            [mock.call(db_amphora1_mock, timeout_dict=self.timeout_dict),
             mock.call(db_amphora2_mock, timeout_dict=self.timeout_dict)])
        self.assertFalse(
            ret[amphora1_mock[constants.ID]][constants.UNREACHABLE])
        self.assertTrue(
            ret[amphora2_mock[constants.ID]][constants.UNREACHABLE])

    def test_set_amphora_firewall_rules(self,
                                        mock_driver,
                                        mock_generate_uuid,
                                        mock_log,
                                        mock_get_session,
                                        mock_listener_repo_get,
                                        mock_listener_repo_update,
                                        mock_amphora_repo_get,
                                        mock_amphora_repo_update):
        amphora = {constants.ID: AMP_ID, constants.VRRP_IP: '192.0.2.88'}
        mock_amphora_repo_get.return_value = _db_amphora_mock

        set_amp_fw_rules = amphora_driver_tasks.SetAmphoraFirewallRules()

        # Test non-SRIOV VIP path
        set_amp_fw_rules.execute([amphora], 0, [{'non-sriov-vip': True}], {},
                                 timeout_dict=None)

        mock_get_session.assert_not_called()
        mock_driver.set_interface_rules.assert_not_called()

        # Test SRIOV VIP path
        set_amp_fw_rules.execute([amphora], 0, [{'fake_rule': True}], {},
                                 timeout_dict=None)

        mock_amphora_repo_get.assert_called_once_with(_session_mock, id=AMP_ID)

        mock_driver.set_interface_rules.assert_called_once_with(
            _db_amphora_mock, '192.0.2.88', [{'fake_rule': True}],
            timeout_dict=None)



#! /usr/bin/env python

#    Copyright 2014 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import errno
import json
import os
import queue
import stat
import time

from oslo_config import cfg
from oslo_log import log as logging

from octavia.amphorae.backends.agent.api_server import util
from octavia.amphorae.backends.health_daemon import health_sender
from octavia.amphorae.backends.utils import haproxy_query
from octavia.amphorae.backends.utils import keepalivedlvs_query


CONF = cfg.CONF
LOG = logging.getLogger(__name__)
SEQ = 0
# MSG_VER is an incrementing integer heartbeat message format version
# this allows for backward compatibility when the amphora-agent is older
# than the controller version and the message format has backwards
# incompatible changes.
#
# ver 1 - Adds UDP listener status when no pool or members are present
# ver 2 - Switch to all listeners in a single combined haproxy config
# ver 3 - Switch stats reporting to deltas

MSG_VER = 3

DELTA_METRICS = ('bin', 'bout', 'ereq', 'stot')

# Filesystem persistent counters for statistics deltas
COUNTERS = None
COUNTERS_FILE = None


def get_counters_file():
    global COUNTERS_FILE
    if COUNTERS_FILE is None:
        stats_file_path = os.path.join(
            CONF.haproxy_amphora.base_path, "stats_counters.json")
        # Open for read+write and create if necessary
        flags = os.O_RDWR | os.O_CREAT
        # mode 00644
        mode = stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP
        try:
            COUNTERS_FILE = os.fdopen(
                os.open(stats_file_path, flags, mode), 'r+')
        except OSError:
            LOG.info("Failed to open `%s`, ignoring...", stats_file_path)
    COUNTERS_FILE.seek(0)
    return COUNTERS_FILE


def get_counters():
    global COUNTERS
    if COUNTERS is None:
        try:
            COUNTERS = json.load(get_counters_file()) or {}
        except (json.JSONDecodeError, AttributeError):
            COUNTERS = {}
    return COUNTERS


def persist_counters():
    """Attempt to persist the latest statistics values"""
    if COUNTERS is None:
        return
    try:
        stats = json.dumps(COUNTERS)
        counters_file = get_counters_file()
        counters_file.truncate(0)
        counters_file.write(stats)
        counters_file.flush()
    except (OSError, AttributeError):
        LOG.warning("Couldn't persist statistics counter file!")


def list_sock_stat_files(hadir=None):
    stat_sock_files = {}
    if hadir is None:
        hadir = CONF.haproxy_amphora.base_path
    lb_ids = util.get_loadbalancers()
    for lb_id in lb_ids:
        sock_file = lb_id + ".sock"
        stat_sock_files[lb_id] = os.path.join(hadir, sock_file)
    return stat_sock_files


def run_sender(cmd_queue):
    LOG.info('Health Manager Sender starting.')
    sender = health_sender.UDPStatusSender()

    keepalived_cfg_path = util.keepalived_cfg_path()
    keepalived_pid_path = util.keepalived_pid_path()

    while True:
        try:
            # If the keepalived config file is present check
            # that it is running, otherwise don't send the health
            # heartbeat
            if os.path.isfile(keepalived_cfg_path):
                # Is there a pid file for keepalived?
                with open(keepalived_pid_path, encoding='utf-8') as pid_file:
                    pid = int(pid_file.readline())
                os.kill(pid, 0)

            message = build_stats_message()
            sender.dosend(message)
        except OSError as e:
            if e.errno == errno.ENOENT:
                # Missing PID file, skip health heartbeat.
                LOG.error('Missing keepalived PID file %s, skipping health '
                          'heartbeat.', keepalived_pid_path)
            elif e.errno == errno.ESRCH:
                # Keepalived is not running, skip health heartbeat.
                LOG.error('Keepalived is configured but not running, '
                          'skipping health heartbeat.')
            else:
                LOG.exception('Failed to check keepalived and haproxy status '
                              'due to exception %s, skipping health '
                              'heartbeat.', str(e))
        except Exception as e:
            LOG.exception('Failed to check keepalived and haproxy status due '
                          'to exception %s, skipping health heartbeat.',
                          str(e))

        try:
            cmd = cmd_queue.get_nowait()
            if cmd == 'reload':
                LOG.info('Reloading configuration')
                CONF.reload_config_files()
            elif cmd == 'shutdown':
                LOG.info('Health Manager Sender shutting down.')
                break
        except queue.Empty:
            pass
        time.sleep(CONF.health_manager.heartbeat_interval)


def get_stats(stat_sock_file):
    try:
        stats_query = haproxy_query.HAProxyQuery(stat_sock_file)
        stats = stats_query.show_stat()
        pool_status = stats_query.get_pool_status()
    except Exception as e:
        LOG.warning('Unable to query the HAProxy stats (%s) due to: %s',
                    stat_sock_file, str(e))
        # Return empty lists so that the heartbeat will still be sent
        return [], {}
    return stats, pool_status


def calculate_stats_deltas(listener_id, row):
    counters = get_counters()
    listener_counters = counters.get(listener_id, {})
    counters[listener_id] = listener_counters

    delta_values = {}
    for metric_key in DELTA_METRICS:
        current_value = int(row[metric_key])
        # Get existing counter for our metrics
        last_value = listener_counters.get(metric_key, 0)
        # Store the new absolute value
        listener_counters[metric_key] = current_value
        # Calculate a delta for each metric
        delta = current_value - last_value
        # Did HAProxy restart or reset counters?
        if delta < 0:
            delta = current_value  # If so, reset ours.
        delta_values[metric_key] = delta

    return delta_values


def build_stats_message():
    """Build a stats message based on retrieved listener statistics.

    Example version 3 message without UDP (note that values are deltas,
    not absolutes)::

        {"id": "<amphora_id>",
         "seq": 67,
         "listeners": {
           "<listener_id>": {
             "status": "OPEN",
             "stats": {
               "tx": 0,
               "rx": 0,
               "conns": 0,
               "totconns": 0,
               "ereq": 0
             }
           }
         },
         "pools": {
             "<pool_id>:<listener_id>": {
               "status": "UP",
               "members": {
                 "<member_id>": "no check"
               }
             }
         },
         "ver": 3
        }
    """
    global SEQ
    msg = {'id': CONF.amphora_agent.amphora_id,
           'seq': SEQ, 'listeners': {}, 'pools': {},
           'ver': MSG_VER}
    SEQ += 1
    stat_sock_files = list_sock_stat_files()
    # TODO(rm_work) There should only be one of these in the new config system
    for lb_id, stat_sock_file in stat_sock_files.items():
        if util.is_lb_running(lb_id):
            (stats, pool_status) = get_stats(stat_sock_file)
            for row in stats:
                if row['svname'] == 'FRONTEND':
                    listener_id = row['pxname']
                    delta_values = calculate_stats_deltas(listener_id, row)
                    msg['listeners'][listener_id] = {
                        'status': row['status'],
                        'stats': {'tx': delta_values['bout'],
                                  'rx': delta_values['bin'],
                                  'conns': int(row['scur']),
                                  'totconns': delta_values['stot'],
                                  'ereq': delta_values['ereq']}}
            for pool_id, pool in pool_status.items():
                msg['pools'][pool_id] = {"status": pool['status'],
                                         "members": pool['members']}

    # UDP listener part
    lvs_listener_ids = util.get_lvs_listeners()
    if lvs_listener_ids:
        listeners_stats = keepalivedlvs_query.get_lvs_listeners_stats()
        if listeners_stats:
            for listener_id, listener_stats in listeners_stats.items():
                delta_values = calculate_stats_deltas(
                    listener_id, listener_stats['stats'])
                pool_status = (
                    keepalivedlvs_query.get_lvs_listener_pool_status(
                        listener_id))
                lvs_listener_dict = {}
                lvs_listener_dict['status'] = listener_stats['status']
                lvs_listener_dict['stats'] = {
                    'tx': delta_values['bout'],
                    'rx': delta_values['bin'],
                    'conns': listener_stats['stats']['scur'],
                    'totconns': delta_values['stot'],
                    'ereq': delta_values['ereq']
                }
                if pool_status:
                    pool_id = pool_status['lvs']['uuid']
                    msg['pools'][pool_id] = {
                        "status": pool_status['lvs']['status'],
                        "members": pool_status['lvs']['members']
                    }
                msg['listeners'][listener_id] = lvs_listener_dict
    persist_counters()
    return msg



# Copyright 2018 Rackspace US Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from octavia.common import constants as consts

# This is a JSON schema validation dictionary
# https://json-schema.org/latest/json-schema-validation.html
#
# Note: This is used to generate the amphora driver "supported flavor
#       metadata" dictionary. Each property should include a description
#       for the user to understand what this flavor setting does.
#
# Where possible, the property name should match the configuration file name
# for the setting. The configuration file setting is the default when a
# setting is not defined in a flavor profile.

SUPPORTED_FLAVOR_SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "title": "Octavia Amphora Driver Flavor Metadata Schema",
    "description": "This schema is used to validate new flavor profiles "
                   "submitted for use in an amphora driver flavor profile.",
    "type": "object",
    "additionalProperties": False,
    "properties": {
        consts.LOADBALANCER_TOPOLOGY: {
            "type": "string",
            "description": "The load balancer topology. One of: "
                           "SINGLE - One amphora per load balancer. "
                           "ACTIVE_STANDBY - Two amphora per load balancer.",
            "enum": list(consts.SUPPORTED_LB_TOPOLOGIES)
        },
        consts.COMPUTE_FLAVOR: {
            "type": "string",
            "description": "The compute driver flavor ID."
        },
        consts.AMP_IMAGE_TAG: {
            "type": "string",
            "description": "The amphora image tag."
        },
        consts.SRIOV_VIP: {
            "type": "boolean",
            "description": "When true, the VIP port will be created using an "
                           "SR-IOV VF port."
        },
        consts.ALLOW_MEMBER_SRIOV: {
            "type": "boolean",
            "description": "When true, users can request a member port be "
                           "SR-IOV enabled at member creation time."
        }
    }
}



#    Copyright 2014 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from pecan import hooks
from webob import acceptparse
from webob import exc

from octavia.api.common import pagination
from octavia.api.common import utils
from octavia.common import constants
from octavia.common import context
from octavia.i18n import _

_HEALTHCHECK_PATHS = ['/healthcheck', '/load-balancer/healthcheck']


class ContentTypeHook(hooks.PecanHook):
    """Force the request content type to JSON if that is acceptable."""

    def on_route(self, state):
        # Oslo healthcheck middleware has its own content type handling
        # so we need to bypass the Octavia content type restrictions.
        if state.request.path in _HEALTHCHECK_PATHS:
            return
        # TODO(johnsom) Testing for an empty string is a workaround for an
        #               openstacksdk bug present up to the initial
        #               antelope release of openstacksdk. This means the
        #               octavia dashboard would also be impacted.
        #               This can be removed once antelope is EOL.
        # See: https://review.opendev.org/c/openstack/openstacksdk/+/876669
        if state.request.accept and state.request.accept.header_value != '':
            best_matches = state.request.accept.acceptable_offers(
                [constants.APPLICATION_JSON])
            if not best_matches:
                # The API reference says we always respond with JSON
                state.request.accept = acceptparse.create_accept_header(
                    constants.APPLICATION_JSON)
                msg = _('Only content type %s is accepted.')
                raise exc.HTTPNotAcceptable(
                    msg % constants.APPLICATION_JSON,
                    json_formatter=utils.json_error_formatter)

        # Force application/json with no other options for the request
        state.request.accept = acceptparse.create_accept_header(
            constants.APPLICATION_JSON)


class ContextHook(hooks.PecanHook):
    """Configures a request context and attaches it to the request."""

    def on_route(self, state):
        context_obj = context.RequestContext.from_environ(
            state.request.environ)
        state.request.context['octavia_context'] = context_obj


class QueryParametersHook(hooks.PecanHook):

    def before(self, state):
        if state.request.method != 'GET':
            return

        state.request.context[
            constants.PAGINATION_HELPER] = pagination.PaginationHelper(
            state.request.params.mixed())



# Copyright 2024 Red Hat
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
from taskflow import task

from octavia.common import constants


class AmphoraToAmphoraeWithVRRPIP(task.Task):
    """A shim class to convert a single Amphora instance to a list."""

    def execute(self, amphora: dict, base_port: dict):
        # The VRRP_IP has not been stamped on the Amphora at this point in the
        # flow, so inject it from our port create call in a previous task.
        amphora[constants.VRRP_IP] = (
            base_port[constants.FIXED_IPS][0][constants.IP_ADDRESS])
        return [amphora]



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
#    under the License.


import os
import re
import stat
import subprocess
import typing as tp

import jinja2
from oslo_config import cfg
from oslo_log import log as logging

from octavia.amphorae.backends.utils import ip_advertisement
from octavia.amphorae.backends.utils import network_utils
from octavia.common import constants as consts

CONF = cfg.CONF
LOG = logging.getLogger(__name__)

FRONTEND_BACKEND_PATTERN = re.compile(r'\n(frontend|backend)\s+(\S+)\n')
LISTENER_MODE_PATTERN = re.compile(r'^\s+mode\s+(.*)$', re.MULTILINE)
TLS_CERT_PATTERN = re.compile(r'^\s+bind\s+\S+\s+ssl crt-list\s+(\S*)',
                              re.MULTILINE)
STATS_SOCKET_PATTERN = re.compile(r'stats socket\s+(\S+)')


class ParsingError(Exception):
    pass


def init_path(lb_id):
    return os.path.join(consts.SYSTEMD_DIR, f'haproxy-{lb_id}.service')


def keepalived_lvs_dir():
    return os.path.join(CONF.haproxy_amphora.base_path, 'lvs')


def keepalived_lvs_init_path(listener_id):
    return os.path.join(consts.SYSTEMD_DIR,
                        consts.KEEPALIVEDLVS_SYSTEMD %
                        str(listener_id))


def keepalived_backend_check_script_dir():
    return os.path.join(CONF.haproxy_amphora.base_path, 'lvs/check/')


def keepalived_backend_check_script_path():
    return os.path.join(keepalived_backend_check_script_dir(),
                        'udp_check.sh')


def keepalived_lvs_pids_path(listener_id):
    pids_path = {}
    for file_ext in ['pid', 'vrrp.pid', 'check.pid']:
        pids_path[file_ext] = (
            os.path.join(CONF.haproxy_amphora.base_path,
                         f"lvs/octavia-keepalivedlvs-{str(listener_id)}."
                         f"{file_ext}"))
    return pids_path['pid'], pids_path['vrrp.pid'], pids_path['check.pid']


def keepalived_lvs_cfg_path(listener_id):
    return os.path.join(CONF.haproxy_amphora.base_path,
                        f"lvs/octavia-keepalivedlvs-{str(listener_id)}.conf")


def haproxy_dir(lb_id):
    return os.path.join(CONF.haproxy_amphora.base_path, lb_id)


def pid_path(lb_id):
    return os.path.join(haproxy_dir(lb_id), lb_id + '.pid')


def config_path(lb_id):
    return os.path.join(haproxy_dir(lb_id), 'haproxy.cfg')


def state_file_path(lb_id):
    return os.path.join(haproxy_dir(lb_id), 'servers-state')


def get_haproxy_pid(lb_id):
    with open(pid_path(lb_id), encoding='utf-8') as f:
        return f.readline().rstrip()


def get_keepalivedlvs_pid(listener_id):
    pid_file = keepalived_lvs_pids_path(listener_id)[0]
    with open(pid_file, encoding='utf-8') as f:
        return f.readline().rstrip()


def haproxy_sock_path(lb_id):
    return os.path.join(CONF.haproxy_amphora.base_path, lb_id + '.sock')


def haproxy_check_script_path():
    return os.path.join(keepalived_check_scripts_dir(),
                        'haproxy_check_script.sh')


def keepalived_dir():
    return os.path.join(CONF.haproxy_amphora.base_path, 'vrrp')


def keepalived_init_path():
    return os.path.join(consts.SYSTEMD_DIR, consts.KEEPALIVED_SYSTEMD)


def keepalived_pid_path():
    return os.path.join(CONF.haproxy_amphora.base_path,
                        'vrrp/octavia-keepalived.pid')


def keepalived_cfg_path():
    return os.path.join(CONF.haproxy_amphora.base_path,
                        'vrrp/octavia-keepalived.conf')


def keepalived_log_path():
    return os.path.join(CONF.haproxy_amphora.base_path,
                        'vrrp/octavia-keepalived.log')


def keepalived_check_scripts_dir():
    return os.path.join(CONF.haproxy_amphora.base_path,
                        'vrrp/check_scripts')


def keepalived_check_script_path():
    return os.path.join(CONF.haproxy_amphora.base_path,
                        'vrrp/check_script.sh')


def get_listeners():
    """Get Listeners

    :returns: An array with the ids of all listeners, e.g. ['123', '456', ...]
              or [] if no listeners exist
    """
    listeners = []
    for lb_id in get_loadbalancers():
        listeners_on_lb = parse_haproxy_file(lb_id)[1]
        listeners.extend(list(listeners_on_lb.keys()))
    return listeners


def get_loadbalancers():
    """Get Load balancers

    :returns: An array with the uuids of all load balancers,
              e.g. ['123', '456', ...] or [] if no loadbalancers exist
    """
    if os.path.exists(CONF.haproxy_amphora.base_path):
        return [f for f in os.listdir(CONF.haproxy_amphora.base_path)
                if os.path.exists(config_path(f))]
    return []


def is_lb_running(lb_id):
    return os.path.exists(pid_path(lb_id)) and os.path.exists(
        os.path.join('/proc', get_haproxy_pid(lb_id)))


def get_lvs_listeners():
    result = []
    if os.path.exists(keepalived_lvs_dir()):
        for f in os.listdir(keepalived_lvs_dir()):
            if f.endswith('.conf'):
                prefix = f.split('.')[0]
                if re.search("octavia-keepalivedlvs-", prefix):
                    result.append(f.split(
                        'octavia-keepalivedlvs-')[1].split('.')[0])
    return result


def is_lvs_listener_running(listener_id):
    pid_file = keepalived_lvs_pids_path(listener_id)[0]
    return os.path.exists(pid_file) and os.path.exists(
        os.path.join('/proc', get_keepalivedlvs_pid(listener_id)))


def install_netns_systemd_service():
    flags = os.O_WRONLY | os.O_CREAT | os.O_TRUNC
    # mode 00644
    mode = stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH

    # TODO(bcafarel): implement this for other init systems
    # netns handling depends on a separate unit file
    netns_path = os.path.join(consts.SYSTEMD_DIR,
                              consts.AMP_NETNS_SVC_PREFIX + '.service')

    jinja_env = jinja2.Environment(
        autoescape=True, loader=jinja2.FileSystemLoader(os.path.dirname(
            os.path.realpath(__file__)
        ) + consts.AGENT_API_TEMPLATES))

    if not os.path.exists(netns_path):
        with os.fdopen(os.open(netns_path, flags, mode), 'w') as text_file:
            text = jinja_env.get_template(
                consts.AMP_NETNS_SVC_PREFIX + '.systemd.j2').render(
                    amphora_nsname=consts.AMPHORA_NAMESPACE)
            text_file.write(text)


def run_systemctl_command(command, service, raise_error=True):
    cmd = f"systemctl {command} {service}"
    try:
        subprocess.check_output(cmd.split(), stderr=subprocess.STDOUT,
                                encoding='utf-8')
    except subprocess.CalledProcessError as e:
        LOG.debug("Failed to %(cmd)s %(srvc)s service: "
                  "%(err)s %(out)s", {'cmd': command, 'srvc': service,
                                      'err': e, 'out': e.output})
        if raise_error:
            raise


def get_backend_for_lb_object(object_id):
    """Returns the backend for a listener.

    If the listener is a TCP based listener return 'HAPROXY'.
    If the listener is a UDP or SCTP based listener return 'LVS'
    If the listener is not identifiable, return None.

    :param listener_id: The ID of the listener to identify.
    :returns: HAPROXY_BACKEND, LVS_BACKEND or None
    """
    if os.path.exists(config_path(object_id)):
        return consts.HAPROXY_BACKEND
    if os.path.exists(keepalived_lvs_cfg_path(object_id)):
        return consts.LVS_BACKEND
    return None


def parse_haproxy_file(lb_id):
    with open(config_path(lb_id), encoding='utf-8') as file:
        cfg = file.read()

        listeners = {}

        m = FRONTEND_BACKEND_PATTERN.split(cfg)
        last_token = None
        last_id = None
        for section in m:
            if last_token is None:
                # We aren't in a section yet, see if this line starts one
                if section == 'frontend':
                    last_token = section
            elif last_token == 'frontend':
                # We're in a frontend section, save the id for later
                last_token = last_token + "_id"
                last_id = section
            elif last_token == 'frontend_id':
                # We're in a frontend section and already have the id
                # Look for the mode
                mode_matcher = LISTENER_MODE_PATTERN.search(section)
                if not mode_matcher:
                    raise ParsingError()
                listeners[last_id] = {
                    'mode': mode_matcher.group(1).upper(),
                }
                # Now see if this is a TLS frontend
                tls_matcher = TLS_CERT_PATTERN.search(section)
                if tls_matcher:
                    # TODO(rm_work): Can't we have terminated tcp?
                    listeners[last_id]['mode'] = 'TERMINATED_HTTPS'
                    listeners[last_id]['ssl_crt'] = tls_matcher.group(1)
                # Clear out the token and id and start over
                last_token = last_id = None

        m = STATS_SOCKET_PATTERN.search(cfg)
        if not m:
            raise ParsingError()
        stats_socket = m.group(1)

        return stats_socket, listeners


def vrrp_check_script_update(lb_id, action):
    os.makedirs(keepalived_dir(), exist_ok=True)
    os.makedirs(keepalived_check_scripts_dir(), exist_ok=True)

    lb_ids = get_loadbalancers()
    lvs_ids = get_lvs_listeners()
    # If no LBs are found, so make sure keepalived thinks haproxy is down.
    if not lb_ids:
        if not lvs_ids:
            with open(haproxy_check_script_path(),
                      'w', encoding='utf-8') as text_file:
                text_file.write('exit 1')
        else:
            try:
                LOG.debug("Attempting to remove old haproxy check script...")
                os.remove(haproxy_check_script_path())
                LOG.debug("Finished removing old haproxy check script.")
            except FileNotFoundError:
                LOG.debug("No haproxy check script to remove.")
        return
    if action == consts.AMP_ACTION_STOP:
        lb_ids.remove(lb_id)
    args = []
    for lbid in lb_ids:
        args.append(haproxy_sock_path(lbid))

    cmd = f"haproxy-vrrp-check {' '.join(args)}; exit $?"
    with open(haproxy_check_script_path(), 'w', encoding='utf-8') as text_file:
        text_file.write(cmd)


def get_haproxy_vip_addresses(lb_id):
    """Get the VIP addresses for a load balancer.

    :param lb_id: The load balancer ID to get VIP addresses from.
    :returns: List of VIP addresses (IPv4 and IPv6)
    """
    vips = []
    with open(config_path(lb_id), encoding='utf-8') as file:
        for line in file:
            current_line = line.strip()
            if current_line.startswith('bind'):
                for section in current_line.split(' '):
                    # We will always have a port assigned per the template.
                    if ':' in section:
                        if ',' in section:
                            addr_port = section.rstrip(',')
                            vips.append(addr_port.rpartition(':')[0])
                        else:
                            vips.append(section.rpartition(':')[0])
                            break
    return vips


def get_lvs_vip_addresses(listener_id: str) -> list[str]:
    """Get the VIP addresses for a LVS load balancer.

    :param listener_id: The listener ID to get VIP addresses from.
    :returns: List of VIP addresses (IPv4 and IPv6)
    """
    vips = []
    # Extract the VIP addresses from keepalived configuration
    # Format is
    # virtual_server_group ipv<n>-group {
    #     vip_address1 port1
    #     vip_address2 port2
    # }
    # it can be repeated in case of dual-stack LBs
    with open(keepalived_lvs_cfg_path(listener_id), encoding='utf-8') as file:
        vsg_section = False
        for line in file:
            current_line = line.strip()
            if vsg_section:
                if current_line.startswith('}'):
                    vsg_section = False
                else:
                    vip_address = current_line.split(' ')[0]
                    vips.append(vip_address)
            elif line.startswith('virtual_server_group '):
                vsg_section = True
    return vips


def send_vip_advertisements(lb_id: tp.Optional[str] = None,
                            listener_id: tp.Optional[str] = None):
    """Sends address advertisements for each load balancer VIP.

    This method will send either GARP (IPv4) or neighbor advertisements (IPv6)
    for the VIP addresses on a load balancer.

    :param lb_id: The load balancer ID to send advertisements for.
    :returns: None
    """
    try:
        if lb_id:
            vips = get_haproxy_vip_addresses(lb_id)
        else:
            vips = get_lvs_vip_addresses(listener_id)

        for vip in vips:
            interface = network_utils.get_interface_name(
                vip, net_ns=consts.AMPHORA_NAMESPACE)
            ip_advertisement.send_ip_advertisement(
                interface, vip, net_ns=consts.AMPHORA_NAMESPACE)
    except Exception as e:
        LOG.debug('Send VIP advertisement failed due to :%s. '
                  'This amphora may not be the MASTER. Ignoring.', str(e))


def send_member_advertisements(fixed_ips: tp.Iterable[tp.Dict[str, str]]):
    """Sends advertisements for each fixed_ip of a list

    This method will send either GARP (IPv4) or neighbor advertisements (IPv6)
    for the addresses of the subnets of the members.

    :param fixed_ips: a list of dicts that contain 'ip_address' elements
    :returns: None
    """
    try:
        for fixed_ip in fixed_ips:
            ip_address = fixed_ip[consts.IP_ADDRESS]
            interface = network_utils.get_interface_name(
                ip_address, net_ns=consts.AMPHORA_NAMESPACE)
            ip_advertisement.send_ip_advertisement(
                interface, ip_address, net_ns=consts.AMPHORA_NAMESPACE)
    except Exception as e:
        LOG.debug('Send member advertisement failed due to: %s', str(e))



#    Copyright 2016 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from taskflow import task

from octavia.common import constants
from octavia.controller.worker import task_utils as task_utilities


class BaseLifecycleTask(task.Task):
    """Base task to instantiate common classes."""

    def __init__(self, **kwargs):
        self.task_utils = task_utilities.TaskUtils()
        super().__init__(**kwargs)


class AmphoraIDToErrorOnRevertTask(BaseLifecycleTask):
    """Task to checkpoint Amphora lifecycle milestones."""

    def execute(self, amphora_id):
        pass

    def revert(self, amphora_id, *args, **kwargs):
        self.task_utils.mark_amphora_status_error(amphora_id)


class AmphoraToErrorOnRevertTask(AmphoraIDToErrorOnRevertTask):
    """Task to checkpoint Amphora lifecycle milestones."""

    def execute(self, amphora):
        pass

    def revert(self, amphora, *args, **kwargs):
        super().revert(
            amphora.get(constants.ID))


class HealthMonitorToErrorOnRevertTask(BaseLifecycleTask):
    """Task to set a member to ERROR on revert."""

    def execute(self, health_mon, listeners, loadbalancer):
        pass

    def revert(self, health_mon, listeners, loadbalancer, *args, **kwargs):
        try:
            self.task_utils.mark_health_mon_prov_status_error(
                health_mon[constants.HEALTHMONITOR_ID])
            self.task_utils.mark_pool_prov_status_active(
                health_mon[constants.POOL_ID])
            for listener in listeners:
                self.task_utils.mark_listener_prov_status_active(
                    listener[constants.LISTENER_ID])
        except Exception:
            # Catching and skipping, errors are already reported by task_utils
            # and we want to ensure that mark_loadbalancer_prov_status_active
            # is called to unlock the LB (it will pass or it will fail after a
            # very long timeout)
            pass
        self.task_utils.mark_loadbalancer_prov_status_active(
            loadbalancer[constants.LOADBALANCER_ID])


class L7PolicyToErrorOnRevertTask(BaseLifecycleTask):
    """Task to set a l7policy to ERROR on revert."""

    def execute(self, l7policy, listeners, loadbalancer_id):
        pass

    def revert(self, l7policy, listeners, loadbalancer_id, *args, **kwargs):
        try:
            self.task_utils.mark_l7policy_prov_status_error(
                l7policy[constants.L7POLICY_ID])
            for listener in listeners:
                self.task_utils.mark_listener_prov_status_active(
                    listener[constants.LISTENER_ID])
        except Exception:
            # Catching and skipping, errors are already reported by task_utils
            # and we want to ensure that mark_loadbalancer_prov_status_active
            # is called to unlock the LB (it will pass or it will fail after a
            # very long timeout)
            pass
        self.task_utils.mark_loadbalancer_prov_status_active(loadbalancer_id)


class L7RuleToErrorOnRevertTask(BaseLifecycleTask):
    """Task to set a l7rule to ERROR on revert."""

    def execute(self, l7rule, l7policy_id, listeners, loadbalancer_id):
        pass

    def revert(self, l7rule, l7policy_id, listeners, loadbalancer_id, *args,
               **kwargs):
        try:
            self.task_utils.mark_l7rule_prov_status_error(
                l7rule[constants.L7RULE_ID])
            self.task_utils.mark_l7policy_prov_status_active(l7policy_id)
            for listener in listeners:
                self.task_utils.mark_listener_prov_status_active(
                    listener[constants.LISTENER_ID])
        except Exception:
            # Catching and skipping, errors are already reported by task_utils
            # and we want to ensure that mark_loadbalancer_prov_status_active
            # is called to unlock the LB (it will pass or it will fail after a
            # very long timeout)
            pass
        self.task_utils.mark_loadbalancer_prov_status_active(
            loadbalancer_id)


class ListenerToErrorOnRevertTask(BaseLifecycleTask):
    """Task to set a listener to ERROR on revert."""

    def execute(self, listener):
        pass

    def revert(self, listener, *args, **kwargs):
        try:
            self.task_utils.mark_listener_prov_status_error(
                listener[constants.LISTENER_ID])
        except Exception:
            # Catching and skipping, errors are already reported by task_utils
            # and we want to ensure that mark_loadbalancer_prov_status_active
            # is called to unlock the LB (it will pass or it will fail after a
            # very long timeout)
            pass
        self.task_utils.mark_loadbalancer_prov_status_active(
            listener[constants.LOADBALANCER_ID])


class ListenersToErrorOnRevertTask(BaseLifecycleTask):
    """Task to set a listener to ERROR on revert."""

    def execute(self, listeners):
        pass

    def revert(self, listeners, *args, **kwargs):
        try:
            for listener in listeners:
                self.task_utils.mark_listener_prov_status_error(
                    listener[constants.LISTENER_ID])
        except Exception:
            # Catching and skipping, errors are already reported by task_utils
            # and we want to ensure that mark_loadbalancer_prov_status_active
            # is called to unlock the LB (it will pass or it will fail after a
            # very long timeout)
            pass
        self.task_utils.mark_loadbalancer_prov_status_active(
            listeners[0][constants.LOADBALANCER_ID])


class LoadBalancerIDToErrorOnRevertTask(BaseLifecycleTask):
    """Task to set the load balancer to ERROR on revert."""

    def execute(self, loadbalancer_id):
        pass

    def revert(self, loadbalancer_id, *args, **kwargs):
        self.task_utils.mark_loadbalancer_prov_status_error(loadbalancer_id)


class LoadBalancerToErrorOnRevertTask(LoadBalancerIDToErrorOnRevertTask):
    """Task to set the load balancer to ERROR on revert."""

    def execute(self, loadbalancer):
        pass

    def revert(self, loadbalancer, *args, **kwargs):
        super().revert(
            loadbalancer[constants.LOADBALANCER_ID])


class MemberToErrorOnRevertTask(BaseLifecycleTask):
    """Task to set a member to ERROR on revert."""

    def execute(self, member, listeners, loadbalancer, pool_id):
        pass

    def revert(self, member, listeners, loadbalancer, pool_id, *args,
               **kwargs):
        try:
            self.task_utils.mark_member_prov_status_error(
                member[constants.MEMBER_ID])
            for listener in listeners:
                self.task_utils.mark_listener_prov_status_active(
                    listener[constants.LISTENER_ID])
            self.task_utils.mark_pool_prov_status_active(pool_id)
        except Exception:
            # Catching and skipping, errors are already reported by task_utils
            # and we want to ensure that mark_loadbalancer_prov_status_active
            # is called to unlock the LB (it will pass or it will fail after a
            # very long timeout)
            pass
        self.task_utils.mark_loadbalancer_prov_status_active(
            loadbalancer[constants.LOADBALANCER_ID])


class MembersToErrorOnRevertTask(BaseLifecycleTask):
    """Task to set members to ERROR on revert."""

    def execute(self, members, listeners, loadbalancer, pool_id):
        pass

    def revert(self, members, listeners, loadbalancer, pool_id, *args,
               **kwargs):
        try:
            for m in members:
                self.task_utils.mark_member_prov_status_error(
                    m[constants.MEMBER_ID])
            for listener in listeners:
                self.task_utils.mark_listener_prov_status_active(
                    listener[constants.LISTENER_ID])
            self.task_utils.mark_pool_prov_status_active(pool_id)
        except Exception:
            # Catching and skipping, errors are already reported by task_utils
            # and we want to ensure that mark_loadbalancer_prov_status_active
            # is called to unlock the LB (it will pass or it will fail after a
            # very long timeout)
            pass
        self.task_utils.mark_loadbalancer_prov_status_active(
            loadbalancer[constants.LOADBALANCER_ID])


class PoolToErrorOnRevertTask(BaseLifecycleTask):
    """Task to set a pool to ERROR on revert."""

    def execute(self, pool_id, listeners, loadbalancer):
        pass

    def revert(self, pool_id, listeners, loadbalancer, *args, **kwargs):
        try:
            self.task_utils.mark_pool_prov_status_error(pool_id)
            for listener in listeners:
                self.task_utils.mark_listener_prov_status_active(
                    listener[constants.LISTENER_ID])
        except Exception:
            # Catching and skipping, errors are already reported by task_utils
            # and we want to ensure that mark_loadbalancer_prov_status_active
            # is called to unlock the LB (it will pass or it will fail after a
            # very long timeout)
            pass
        self.task_utils.mark_loadbalancer_prov_status_active(
            loadbalancer[constants.LOADBALANCER_ID])



#    Copyright 2015 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from keystoneauth1 import exceptions as ks_exceptions
from keystoneauth1 import loading as ks_loading
from keystonemiddleware import auth_token
from oslo_config import cfg
from oslo_log import log as logging

from octavia.common import constants

LOG = logging.getLogger(__name__)

_NOAUTH_PATHS = ['/', '/load-balancer/', '/healthcheck',
                 '/load-balancer/healthcheck']


class KeystoneSession:

    def __init__(self, section=constants.SERVICE_AUTH):
        self._session = None
        self._auth = None

        self.section = section

    def get_session(self, auth=None):
        """Initializes a Keystone session.

        :return: a Keystone Session object
        """
        if auth:
            # Do not use the singleton with custom auth params
            return ks_loading.load_session_from_conf_options(
                cfg.CONF, self.section, auth=auth)

        if not self._session:
            self._session = ks_loading.load_session_from_conf_options(
                cfg.CONF, self.section, auth=self.get_auth())

        return self._session

    def get_auth(self):
        if not self._auth:
            try:
                self._auth = ks_loading.load_auth_from_conf_options(
                    cfg.CONF, self.section)
            except ks_exceptions.auth_plugins.MissingRequiredOptions as e:
                if self.section == constants.SERVICE_AUTH:
                    raise e
                # NOTE(gthiemonge): MissingRequiredOptions is raised: there is
                # one or more missing auth options in the config file. It may
                # be due to the migration from python-neutronclient to
                # openstacksdk.
                # With neutronclient, most of the auth settings were in
                # [service_auth] with a few overrides in [neutron],
                # but with openstacksdk, we have all the auth settings in the
                # [neutron] section. In order to support smooth upgrades, in
                # case those options are missing, we override the undefined
                # options with the existing settings from [service_auth].

                # This code should be removed when all the deployment tools set
                # the correct options in [neutron]

                # The config options are lazily registered/loaded by keystone,
                # it means that we cannot get/set them before invoking
                # 'load_auth_from_conf_options' on 'service_auth'.
                ks_loading.load_auth_from_conf_options(
                    cfg.CONF, constants.SERVICE_AUTH)

                config = getattr(cfg.CONF, self.section)
                for opt in config:
                    # For each option in the [section] section, get its setting
                    # location, if the location is 'opt_default', it means that
                    # the option is not configured in the config file.
                    # if the option is also defined in [service_auth], the
                    # option of the [section] can be replaced by the one from
                    # [service_auth]
                    loc = cfg.CONF.get_location(opt, self.section)
                    if not loc or loc.location == cfg.Locations.opt_default:
                        if hasattr(cfg.CONF.service_auth, opt):
                            cur_value = getattr(config, opt)
                            value = getattr(cfg.CONF.service_auth, opt)
                            if value != cur_value:
                                log_value = (value if opt != "password"
                                             else "<hidden>")
                                LOG.debug("Overriding [%s].%s with '%s'",
                                          self.section, opt, log_value)
                                cfg.CONF.set_override(opt, value, self.section)

                # Now we can call load_auth_from_conf_options for this specific
                # service with the newly defined options.
                self._auth = ks_loading.load_auth_from_conf_options(
                    cfg.CONF, self.section)

        return self._auth

    def get_service_user_id(self):
        return self.get_auth().get_user_id(self.get_session())


class SkippingAuthProtocol(auth_token.AuthProtocol):
    """SkippingAuthProtocol to reach special endpoints

    Bypasses keystone authentication for special request paths, such
    as the api version discovery path.

    Note:
        SkippingAuthProtocol is lean customization
        of :py:class:`keystonemiddleware.auth_token.AuthProtocol`
        that disables keystone communication if the request path
        is in the _NOAUTH_PATHS list.

    """

    def process_request(self, request):
        path = request.path
        if path in _NOAUTH_PATHS:
            LOG.debug('Request path is %s and it does not require keystone '
                      'authentication', path)
            return None  # return NONE to reach actual logic

        return super().process_request(request)



# Copyright 2020 Red Hat, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Add PROXY v2 pool protocol

Revision ID: e6ee84f0abf3
Revises: 2ab994dd3ec2
Create Date: 2020-08-24 11:12:46.745185

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy import sql


# revision identifiers, used by Alembic.
revision = 'e6ee84f0abf3'
down_revision = '2ab994dd3ec2'


def upgrade():
    insert_table = sql.table(
        'protocol',
        sql.column('name', sa.String),
        sql.column('description', sa.String)
    )

    op.bulk_insert(
        insert_table,
        [
            {'name': 'PROXYV2'}
        ]
    )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#

from unittest import mock
import octavia # noqa H306
from octavia.common import constants
from octavia.controller.worker.v2.tasks import notification_tasks
import octavia.tests.unit.base as base


class MockNOTIFIER(mock.MagicMock):
    info = mock.MagicMock()


@mock.patch('octavia.common.rpc.NOTIFIER',
            new_callable=MockNOTIFIER)
@mock.patch('octavia.common.context.RequestContext',
            new_callable=mock.MagicMock)
@mock.patch('octavia.api.v2.types.load_balancer.LoadBalancerFullResponse.'
            'from_data_model',
            new_callable=mock.MagicMock)
class TestNotificationTasks(base.TestCase):
    def test_update_notification_execute(self, *args):
        noti = notification_tasks.SendUpdateNotification()
        id = 1
        lb = {constants.PROJECT_ID: id,
              constants.LOADBALANCER_ID: id}
        noti.execute(lb)
        octavia.common.context.RequestContext.assert_called_with(project_id=id)
        call_args, call_kwargs = octavia.common.rpc.NOTIFIER.info.call_args
        self.assertEqual('octavia.loadbalancer.update.end', call_args[1])

    def test_create_notification(self, *args):
        noti = notification_tasks.SendCreateNotification()
        id = 2
        lb = {constants.PROJECT_ID: id,
              constants.LOADBALANCER_ID: id}
        noti.execute(lb)
        octavia.common.context.RequestContext.assert_called_with(project_id=id)
        call_args, call_kwargs = octavia.common.rpc.NOTIFIER.info.call_args
        self.assertEqual('octavia.loadbalancer.create.end', call_args[1])

    def test_delete_notification(self, *args):
        noti = notification_tasks.SendDeleteNotification()
        id = 3
        lb = {constants.PROJECT_ID: id,
              constants.LOADBALANCER_ID: id}
        noti.execute(lb)
        octavia.common.context.RequestContext.assert_called_with(project_id=id)
        call_args, call_kwargs = octavia.common.rpc.NOTIFIER.info.call_args
        self.assertEqual('octavia.loadbalancer.delete.end', call_args[1])



#    Copyright 2015 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import typing

from openstack.connection import Connection
import openstack.exceptions as os_exceptions
from openstack.network.v2._proxy import Proxy
from oslo_config import cfg
from oslo_log import log as logging

from octavia.common import clients
from octavia.common import data_models
from octavia.i18n import _
from octavia.network import base
from octavia.network import data_models as network_models
from octavia.network.drivers.neutron import utils

if typing.TYPE_CHECKING:
    from octavia.common import context

LOG = logging.getLogger(__name__)
DNS_INT_EXT_ALIAS = 'dns-integration'
SEC_GRP_EXT_ALIAS = 'security-group'
QOS_EXT_ALIAS = 'qos'
CONF_GROUP = 'neutron'

CONF = cfg.CONF


class BaseNeutronDriver(base.AbstractNetworkDriver):
    def __init__(self):
        self.network_proxy: Proxy = self.os_connection.network
        self._check_extension_cache = {}
        self.sec_grp_enabled = self._check_extension_enabled(SEC_GRP_EXT_ALIAS)
        self.dns_integration_enabled = self._check_extension_enabled(
            DNS_INT_EXT_ALIAS)
        self._qos_enabled = self._check_extension_enabled(QOS_EXT_ALIAS)
        self.project_id = self.os_connection.current_project_id

    @property
    def os_connection(self) -> Connection:
        return clients.NeutronAuth.get_neutron_client()

    def _check_extension_enabled(self, extension_alias):
        if extension_alias in self._check_extension_cache:
            status = self._check_extension_cache[extension_alias]
            LOG.debug('Neutron extension %(ext)s cached as %(status)s',
                      {
                          'ext': extension_alias,
                          'status': 'enabled' if status else 'disabled'
                      })
        else:
            if self.network_proxy.find_extension(extension_alias):
                LOG.debug('Neutron extension %(ext)s found enabled',
                          {'ext': extension_alias})
                self._check_extension_cache[extension_alias] = True
            else:
                LOG.debug('Neutron extension %(ext)s is not enabled',
                          {'ext': extension_alias})
                self._check_extension_cache[extension_alias] = False
        return self._check_extension_cache[extension_alias]

    def _port_to_vip(self, port, load_balancer, octavia_owned=False):
        fixed_ip = None
        additional_ips = []
        for port_fixed_ip in port.fixed_ips:
            if (not fixed_ip and
                    port_fixed_ip.subnet_id == load_balancer.vip.subnet_id):
                fixed_ip = port_fixed_ip
            else:
                additional_ips.append(port_fixed_ip)
        kwargs = {
            'ip_address': None,
            'subnet_id': None
        }
        if fixed_ip:
            kwargs['ip_address'] = fixed_ip.ip_address
            kwargs['subnet_id'] = fixed_ip.subnet_id

        primary_vip = data_models.Vip(
            network_id=port.network_id,
            port_id=port.id,
            load_balancer=load_balancer,
            load_balancer_id=load_balancer.id,
            octavia_owned=octavia_owned,
            sg_ids=load_balancer.vip.sg_ids,
            **kwargs)
        additional_vips = [
            data_models.AdditionalVip(
                ip_address=add_fixed_ip.ip_address,
                subnet_id=add_fixed_ip.subnet_id,
                network_id=port.network_id,
                port_id=port.id,
                load_balancer=load_balancer,
                load_balancer_id=load_balancer.id)
            for add_fixed_ip in additional_ips]
        return primary_vip, additional_vips

    def _nova_interface_to_octavia_interface(self, compute_id, nova_interface):
        fixed_ips = [utils.convert_fixed_ip_dict_to_model(fixed_ip)
                     for fixed_ip in nova_interface.fixed_ips]
        return network_models.Interface(compute_id=compute_id,
                                        network_id=nova_interface.net_id,
                                        port_id=nova_interface.port_id,
                                        fixed_ips=fixed_ips)

    def _port_to_octavia_interface(self, compute_id, port):
        fixed_ips = [utils.convert_fixed_ip_dict_to_model(fixed_ip)
                     for fixed_ip in port.get('fixed_ips', [])]
        return network_models.Interface(compute_id=compute_id,
                                        network_id=port['network_id'],
                                        port_id=port['id'],
                                        fixed_ips=fixed_ips)

    def _add_allowed_address_pairs_to_port(self, port_id, ip_address_list):
        aap = [{'ip_address': ip} for ip in ip_address_list]
        self.network_proxy.update_port(port_id,
                                       allowed_address_pairs=aap)

    def _update_security_groups(self, sec_grp_ids: list[str],
                                port_id: str):
        # Note: Neutron accepts the SG even if it already exists
        try:
            self.network_proxy.update_port(
                port_id, security_groups=sec_grp_ids)
        except os_exceptions.NotFoundException as e:
            raise base.PortNotFound(str(e))
        except Exception as e:
            raise base.NetworkException(str(e))

    def _get_ports_by_security_group(self, sec_grp_id):
        all_ports = self.network_proxy.ports(project_id=self.project_id)
        filtered_ports = [
            p for p in all_ports if (p.security_group_ids and
                                     sec_grp_id in p.security_group_ids)]
        return filtered_ports

    def _create_security_group(self, name):
        sec_grp = self.network_proxy.create_security_group(name=name)
        return sec_grp

    def _create_security_group_rule(self, sec_grp_id, protocol,
                                    direction='ingress', port_min=None,
                                    port_max=None, ethertype='IPv6',
                                    cidr=None):
        rule = {
            'security_group_id': sec_grp_id,
            'direction': direction,
            'protocol': protocol,
            'port_range_min': port_min,
            'port_range_max': port_max,
            'ethertype': ethertype,
            'remote_ip_prefix': cidr,
        }

        self.network_proxy.create_security_group_rule(**rule)

    def apply_qos_on_port(self, qos_id, port_id):
        try:
            self.network_proxy.update_port(port_id, qos_policy_id=qos_id)
        except os_exceptions.ResourceNotFound as e:
            raise base.PortNotFound(str(e))
        except Exception as e:
            raise base.NetworkException(str(e))

    def get_plugged_networks(self, compute_id):
        # List neutron ports associated with the Amphora
        try:
            ports = self.network_proxy.ports(device_id=compute_id)
        except Exception:
            LOG.debug('Error retrieving plugged networks for compute '
                      'device %s.', compute_id)
            ports = tuple()
        return [self._port_to_octavia_interface(compute_id, port) for port in
                ports]

    def _get_resource(self, resource_type, resource_id, context=None):
        network = self.network_proxy
        if context and not CONF.networking.allow_invisible_resource_usage:
            network = clients.NeutronAuth.get_user_neutron_client(
                context)

        try:
            resource = getattr(
                network, f"get_{resource_type}")(resource_id)
            return getattr(
                utils, f'convert_{resource_type}_to_model')(resource)
        except os_exceptions.ResourceNotFound as e:
            message = _('{resource_type} not found '
                        '({resource_type} id: {resource_id}).').format(
                resource_type=resource_type, resource_id=resource_id)
            raise getattr(base, '%sNotFound' % ''.join(
                [w.capitalize() for w in resource_type.split('_')]
            ))(message) from e
        except Exception as e:
            message = _('Error retrieving {resource_type} '
                        '({resource_type} id: {resource_id}.').format(
                resource_type=resource_type, resource_id=resource_id)
            LOG.exception(message)
            raise base.NetworkException(message) from e

    def _get_resources_by_filters(self, resource_type, unique_item=False,
                                  **filters):
        """Retrieves item(s) from filters. By default, a list is returned.

        If unique_item set to True, only the first resource is returned.
        """
        try:
            resources = getattr(
                self.network_proxy, f"{resource_type}s")(**filters)
            conversion_function = getattr(
                utils,
                f'convert_{resource_type}_to_model')
            try:
                # get first item to see if there is at least one resource
                res_list = [conversion_function(next(resources))]
            except StopIteration:
                # pylint: disable=raise-missing-from
                raise os_exceptions.NotFoundException(
                    f'No resource of type {resource_type} found that matches '
                    f'given filter criteria: {filters}.')

            if unique_item:
                return res_list[0]
            return res_list + [conversion_function(r) for r in resources]

        except os_exceptions.NotFoundException as e:
            message = _('{resource_type} not found '
                        '({resource_type} Filters: {filters}.').format(
                resource_type=resource_type, filters=filters)
            raise getattr(base, '%sNotFound' % ''.join(
                [w.capitalize() for w in resource_type.split('_')]
            ))(message) from e
        except Exception as e:
            message = _('Error retrieving {resource_type} '
                        '({resource_type} Filters: {filters}.').format(
                resource_type=resource_type, filters=filters)
            LOG.exception(message)
            raise base.NetworkException(message) from e

    def get_network(self, network_id, context=None):
        return self._get_resource('network', network_id, context=context)

    def get_subnet(self, subnet_id, context=None):
        return self._get_resource('subnet', subnet_id, context=context)

    def get_port(self, port_id, context=None):
        return self._get_resource('port', port_id, context=context)

    def get_security_group_by_id(self, sg_id: str,
                                 context: 'context.RequestContext' = None) -> (
            'network_models.SecurityGroup'):
        return self._get_resource('security_group', sg_id, context=context)

    def get_network_by_name(self, network_name):
        return self._get_resources_by_filters(
            'network', unique_item=True, name=network_name)

    def get_subnet_by_name(self, subnet_name):
        return self._get_resources_by_filters(
            'subnet', unique_item=True, name=subnet_name)

    def get_port_by_name(self, port_name):
        return self._get_resources_by_filters(
            'port', unique_item=True, name=port_name)

    def get_port_by_net_id_device_id(self, network_id, device_id):
        return self._get_resources_by_filters(
            'port', unique_item=True,
            network_id=network_id, device_id=device_id)

    def get_qos_policy(self, qos_policy_id):
        return self._get_resource('qos_policy', qos_policy_id)

    def qos_enabled(self):
        return self._qos_enabled

    def get_network_ip_availability(self, network):
        return self._get_resource('network_ip_availability', network.id)

    def plug_fixed_ip(self, port_id, subnet_id, ip_address=None):
        port = self.get_port(port_id).to_dict(recurse=True)
        fixed_ips = port['fixed_ips']

        new_fixed_ip_dict = {'subnet_id': subnet_id}
        if ip_address:
            new_fixed_ip_dict['ip_address'] = ip_address

        fixed_ips.append(new_fixed_ip_dict)

        try:
            updated_port = self.network_proxy.update_port(
                port_id, fixed_ips=fixed_ips)
            return utils.convert_port_to_model(updated_port)
        except Exception as e:
            raise base.NetworkException(str(e))

    def unplug_fixed_ip(self, port_id, subnet_id):
        port = self.get_port(port_id)
        fixed_ips = [
            fixed_ip.to_dict()
            for fixed_ip in port.fixed_ips
            if fixed_ip.subnet_id != subnet_id
        ]

        try:
            updated_port = self.network_proxy.update_port(
                port_id, fixed_ips=fixed_ips)
            return utils.convert_port_to_model(updated_port)
        except Exception as e:
            raise base.NetworkException(str(e))



# Copyright 2015 Hewlett Packard Enterprise Development Company LP
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
from unittest import mock

from oslo_utils import uuidutils

from octavia.amphorae.drivers.keepalived import vrrp_rest_driver
from octavia.common import constants
from octavia.network import data_models as n_data_models
import octavia.tests.unit.base as base

# Version 1.0 is functionally identical to all versions before it
API_VERSION = '1.0'


class TestVRRPRestDriver(base.TestCase):

    def setUp(self):
        self.keepalived_mixin = vrrp_rest_driver.KeepalivedAmphoraDriverMixin()
        self.keepalived_mixin.clients = {
            'base': mock.MagicMock(),
            API_VERSION: mock.MagicMock()}
        self.keepalived_mixin._populate_amphora_api_version = mock.MagicMock()
        self.clients = self.keepalived_mixin.clients
        self.FAKE_CONFIG = 'FAKE CONFIG'
        self.lb_mock = mock.MagicMock()
        self.amphora_mock = mock.MagicMock()
        self.amphora_mock.id = uuidutils.generate_uuid()
        self.amphora_mock.status = constants.AMPHORA_ALLOCATED
        self.amphora_mock.api_version = API_VERSION
        self.lb_mock.amphorae = [self.amphora_mock]
        self.amphorae_network_config = {}
        vip_subnet = mock.MagicMock()
        self.vip_cidr = vip_subnet.cidr = '192.0.2.0/24'
        one_amp_net_config = n_data_models.AmphoraNetworkConfig(
            vip_subnet=vip_subnet
        )
        self.amphorae_network_config[self.amphora_mock.id] = one_amp_net_config

        super().setUp()

    @mock.patch('octavia.amphorae.drivers.keepalived.jinja.'
                'jinja_cfg.KeepalivedJinjaTemplater.build_keepalived_config')
    def test_update_vrrp_conf(self, mock_templater):

        mock_templater.return_value = self.FAKE_CONFIG

        self.keepalived_mixin.update_vrrp_conf(
            self.lb_mock, self.amphorae_network_config, self.amphora_mock)

        mock_templater.assert_called_with(
            self.lb_mock, self.amphora_mock,
            self.amphorae_network_config[self.amphora_mock.id])
        self.clients[API_VERSION].upload_vrrp_config.assert_called_once_with(
            self.amphora_mock,
            self.FAKE_CONFIG)

        # Test with amphorav2 amphorae_network_config list of dicts
        mock_templater.reset_mock()
        self.clients[API_VERSION].upload_vrrp_config.reset_mock()
        v2_amphorae_network_config = {}
        vip_subnet_dict = {
            constants.VIP_SUBNET: {constants.CIDR: '192.0.2.0/24'}}
        v2_amphorae_network_config[self.amphora_mock.id] = vip_subnet_dict

        self.keepalived_mixin.update_vrrp_conf(
            self.lb_mock, v2_amphorae_network_config, self.amphora_mock)

        self.clients[API_VERSION].upload_vrrp_config.assert_called_once_with(
            self.amphora_mock,
            self.FAKE_CONFIG)

        # Test amphora not in AMPHORA_ALLOCATED state
        mock_templater.reset_mock()
        self.clients[API_VERSION].upload_vrrp_config.reset_mock()
        ready_amphora_mock = mock.MagicMock()
        ready_amphora_mock.id = uuidutils.generate_uuid()
        ready_amphora_mock.status = constants.AMPHORA_READY
        ready_amphora_mock.api_version = API_VERSION

        self.keepalived_mixin.update_vrrp_conf(
            self.lb_mock, self.amphorae_network_config, ready_amphora_mock)

        mock_templater.assert_not_called()
        self.clients[API_VERSION].upload_vrrp_config.assert_not_called()

    def test_stop_vrrp_service(self):

        self.keepalived_mixin.stop_vrrp_service(self.lb_mock)

        self.clients[API_VERSION].stop_vrrp.assert_called_once_with(
            self.amphora_mock)

    def test_start_vrrp_service(self):

        self.keepalived_mixin.start_vrrp_service(self.amphora_mock)

        populate_mock = self.keepalived_mixin._populate_amphora_api_version
        populate_mock.assert_called_once_with(self.amphora_mock,
                                              timeout_dict=None)
        self.clients[API_VERSION].start_vrrp.assert_called_once_with(
            self.amphora_mock, timeout_dict=None)

        # Test amphora not in AMPHORA_ALLOCATED state
        self.clients[API_VERSION].start_vrrp.reset_mock()
        ready_amphora_mock = mock.MagicMock()
        ready_amphora_mock.id = uuidutils.generate_uuid()
        ready_amphora_mock.status = constants.AMPHORA_READY
        ready_amphora_mock.api_version = API_VERSION

        self.keepalived_mixin.start_vrrp_service(ready_amphora_mock)

        self.clients[API_VERSION].start_vrrp.assert_not_called()

        # With timeout_dict
        self.clients[API_VERSION].start_vrrp.reset_mock()
        populate_mock.reset_mock()

        timeout_dict = mock.Mock()
        self.keepalived_mixin.start_vrrp_service(self.amphora_mock,
                                                 timeout_dict=timeout_dict)

        populate_mock = self.keepalived_mixin._populate_amphora_api_version
        populate_mock.assert_called_once_with(self.amphora_mock,
                                              timeout_dict=timeout_dict)
        self.clients[API_VERSION].start_vrrp.assert_called_once_with(
            self.amphora_mock, timeout_dict=timeout_dict)

    def test_reload_vrrp_service(self):

        self.keepalived_mixin.reload_vrrp_service(self.lb_mock)

        self.clients[API_VERSION].reload_vrrp.assert_called_once_with(
            self.amphora_mock)



#    Copyright 2014 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

'''initial_create

Revision ID: 35dee79d5865
Revises: None
Create Date: 2014-08-15 11:01:14.897223

'''

from alembic import op
import sqlalchemy as sa
from sqlalchemy import sql

# revision identifiers, used by Alembic.
revision = '35dee79d5865'
down_revision = None


def upgrade():
    # Create lookup tables
    op.create_table(
        'health_monitor_type',
        sa.Column('name', sa.String(30), primary_key=True),
        sa.Column('description', sa.String(255), nullable=True)
    )

    # Create temporary table for table data seeding
    insert_table = sql.table(
        'health_monitor_type',
        sql.column('name', sa.String),
        sql.column('description', sa.String)
    )

    op.bulk_insert(
        insert_table,
        [
            {'name': 'HTTP'},
            {'name': 'HTTPS'},
            {'name': 'TCP'}
        ]
    )

    op.create_table(
        'protocol',
        sa.Column('name', sa.String(30), primary_key=True),
        sa.Column('description', sa.String(255), nullable=True)
    )

    insert_table = sql.table(
        'protocol',
        sql.column('name', sa.String),
        sql.column('description', sa.String)
    )

    op.bulk_insert(
        insert_table,
        [
            {'name': 'HTTP'},
            {'name': 'HTTPS'},
            {'name': 'TCP'}
        ]
    )

    op.create_table(
        'algorithm',
        sa.Column('name', sa.String(30), primary_key=True),
        sa.Column('description', sa.String(255), nullable=True)
    )

    insert_table = sql.table(
        'algorithm',
        sql.column('name', sa.String),
        sql.column('description', sa.String)
    )

    op.bulk_insert(
        insert_table,
        [
            {'name': 'ROUND_ROBIN'},
            {'name': 'LEAST_CONNECTIONS'},
            {'name': 'SOURCE_IP'}
        ]
    )

    op.create_table(
        'session_persistence_type',
        sa.Column('name', sa.String(30), primary_key=True),
        sa.Column('description', sa.String(255), nullable=True)
    )

    insert_table = sql.table(
        'session_persistence_type',
        sql.column('name', sa.String),
        sql.column('description', sa.String)
    )

    op.bulk_insert(
        insert_table,
        [
            {'name': 'SOURCE_IP'},
            {'name': 'HTTP_COOKIE'},
            {'name': 'APP_COOKIE'}
        ]
    )

    op.create_table(
        'provisioning_status',
        sa.Column('name', sa.String(30), primary_key=True),
        sa.Column('description', sa.String(255), nullable=True)
    )

    insert_table = sql.table(
        'provisioning_status',
        sql.column('name', sa.String),
        sql.column('description', sa.String)
    )

    op.bulk_insert(
        insert_table,
        [
            {'name': 'ACTIVE'},
            {'name': 'PENDING_CREATE'},
            {'name': 'PENDING_UPDATE'},
            {'name': 'PENDING_DELETE'},
            {'name': 'DELETED'},
            {'name': 'ERROR'}
        ]
    )

    op.create_table(
        'operating_status',
        sa.Column('name', sa.String(30), primary_key=True),
        sa.Column('description', sa.String(255), nullable=True)
    )

    insert_table = sql.table(
        'operating_status',
        sql.column('name', sa.String),
        sql.column('description', sa.String)
    )

    op.bulk_insert(
        insert_table,
        [
            {'name': 'ONLINE'},
            {'name': 'OFFLINE'},
            {'name': 'DEGRADED'},
            {'name': 'ERROR'}
        ]
    )

    op.create_table(
        'pool',
        sa.Column('tenant_id', sa.String(255), nullable=True),
        sa.Column('id', sa.String(36), nullable=False),
        sa.Column('name', sa.String(255), nullable=True),
        sa.Column('description', sa.String(255), nullable=True),
        sa.Column('protocol', sa.String(16), nullable=False),
        sa.Column('lb_algorithm', sa.String(16), nullable=False),
        sa.Column('operating_status', sa.String(16), nullable=False),
        sa.Column('enabled', sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint('id'),
        sa.ForeignKeyConstraint(['protocol'],
                                ['protocol.name'],
                                name='fk_pool_protocol_name'),
        sa.ForeignKeyConstraint(['lb_algorithm'],
                                ['algorithm.name'],
                                name='fk_pool_algorithm_name'),
        sa.ForeignKeyConstraint(['operating_status'],
                                ['operating_status.name'],
                                name='fk_pool_operating_status_name')
    )

    op.create_table(
        'health_monitor',
        sa.Column('pool_id', sa.String(36), nullable=False),
        sa.Column('type', sa.String(36), nullable=False),
        sa.Column('delay', sa.Integer(), nullable=False),
        sa.Column('timeout', sa.Integer(), nullable=False),
        sa.Column('fall_threshold', sa.Integer(), nullable=False),
        sa.Column('rise_threshold', sa.Integer(), nullable=False),
        sa.Column('http_method', sa.String(16), nullable=True),
        sa.Column('url_path', sa.String(255), nullable=True),
        sa.Column('expected_codes', sa.String(64), nullable=True),
        sa.Column('enabled', sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint('pool_id'),
        sa.ForeignKeyConstraint(['pool_id'], ['pool.id'],
                                name='fk_health_monitor_pool_id'),
        sa.ForeignKeyConstraint(
            ['type'], ['health_monitor_type.name'],
            name='fk_health_monitor_health_monitor_type_name')
    )

    op.create_table(
        'session_persistence',
        sa.Column('pool_id', sa.String(36), nullable=False),
        sa.Column('type', sa.String(16), nullable=False),
        sa.Column('cookie_name', sa.String(255), nullable=True),
        sa.ForeignKeyConstraint(
            ['type'], ['session_persistence_type.name'],
            name='fk_session_persistence_session_persistence_type_name'),
        sa.ForeignKeyConstraint(['pool_id'], ['pool.id'],
                                name='fk_session_persistence_pool_id'),
        sa.PrimaryKeyConstraint('pool_id')
    )

    op.create_table(
        'member',
        sa.Column('tenant_id', sa.String(255), nullable=True),
        sa.Column('id', sa.String(36), nullable=False),
        sa.Column('pool_id', sa.String(36), nullable=False),
        sa.Column('subnet_id', sa.String(36), nullable=True),
        sa.Column('address', sa.String(64), nullable=False),
        sa.Column('protocol_port', sa.Integer(), nullable=False),
        sa.Column('weight', sa.Integer(), nullable=True),
        sa.Column('operating_status', sa.String(16), nullable=False),
        sa.Column('enabled', sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint('id'),
        sa.ForeignKeyConstraint(['pool_id'], ['pool.id'],
                                name='fk_member_pool_id'),
        sa.ForeignKeyConstraint(['operating_status'],
                                ['operating_status.name'],
                                name='fk_member_operating_status_name'),
        sa.UniqueConstraint('pool_id', 'address', 'protocol_port',
                            name='uq_member_pool_id_address_protocol_port')
    )

    op.create_table(
        'load_balancer',
        sa.Column('tenant_id', sa.String(255), nullable=True),
        sa.Column('id', sa.String(36), nullable=False),
        sa.Column('name', sa.String(255), nullable=True),
        sa.Column('description', sa.String(255), nullable=True),
        sa.Column('provisioning_status', sa.String(16), nullable=False),
        sa.Column('operating_status', sa.String(16), nullable=False),
        sa.Column('enabled', sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint('id'),
        sa.ForeignKeyConstraint(
            ['provisioning_status'], ['provisioning_status.name'],
            name='fk_load_balancer_provisioning_status_name'),
        sa.ForeignKeyConstraint(['operating_status'],
                                ['operating_status.name'],
                                name='fk_load_balancer_operating_status_name')
    )

    op.create_table(
        'vip',
        sa.Column('load_balancer_id', sa.String(36), nullable=False),
        sa.Column('ip_address', sa.String(36), nullable=True),
        sa.Column('net_port_id', sa.String(36), nullable=True),
        sa.Column('subnet_id', sa.String(36), nullable=True),
        sa.Column('floating_ip_id', sa.String(36), nullable=True),
        sa.Column('floating_ip_network_id', sa.String(36), nullable=True),
        sa.PrimaryKeyConstraint('load_balancer_id'),
        sa.ForeignKeyConstraint(['load_balancer_id'], ['load_balancer.id'],
                                name='fk_vip_load_balancer_id')
    )

    op.create_table(
        'listener',
        sa.Column('tenant_id', sa.String(255), nullable=True),
        sa.Column('id', sa.String(36), nullable=False),
        sa.Column('name', sa.String(255), nullable=True),
        sa.Column('description', sa.String(255), nullable=True),
        sa.Column('protocol', sa.String(16), nullable=False),
        sa.Column('protocol_port', sa.Integer(), nullable=False),
        sa.Column('connection_limit', sa.Integer(), nullable=True),
        sa.Column('load_balancer_id', sa.String(36), nullable=True),
        sa.Column('tls_certificate_id', sa.String(36), nullable=True),
        sa.Column('default_pool_id', sa.String(36), nullable=True),
        sa.Column('provisioning_status', sa.String(16), nullable=False),
        sa.Column('operating_status', sa.String(16), nullable=False),
        sa.Column('enabled', sa.Boolean(), nullable=False),
        sa.ForeignKeyConstraint(['load_balancer_id'], ['load_balancer.id'],
                                name='fk_listener_load_balancer_id'),
        sa.ForeignKeyConstraint(['default_pool_id'], ['pool.id'],
                                name='fk_listener_pool_id'),
        sa.ForeignKeyConstraint(['protocol'], ['protocol.name'],
                                name='fk_listener_protocol_name'),
        sa.ForeignKeyConstraint(['provisioning_status'],
                                ['provisioning_status.name'],
                                name='fk_listener_provisioning_status_name'),
        sa.ForeignKeyConstraint(['operating_status'],
                                ['operating_status.name'],
                                name='fk_listener_operating_status_name'),
        sa.UniqueConstraint('default_pool_id',
                            name='uq_listener_default_pool_id'),
        sa.UniqueConstraint(
            'load_balancer_id', 'protocol_port',
            name='uq_listener_load_balancer_id_protocol_port'),
        sa.PrimaryKeyConstraint('id')
    )

    op.create_table(
        'sni',
        sa.Column('listener_id', sa.String(36), nullable=False),
        sa.Column('tls_container_id', sa.String(36), nullable=False),
        sa.Column('position', sa.Integer, nullable=True),
        sa.ForeignKeyConstraint(['listener_id'], ['listener.id'],
                                name='fk_sni_listener_id'),
        sa.PrimaryKeyConstraint('listener_id', 'tls_container_id')
    )

    op.create_table(
        'listener_statistics',
        sa.Column('listener_id', sa.String(36), nullable=False),
        sa.Column('bytes_in', sa.BigInteger(), nullable=False),
        sa.Column('bytes_out', sa.BigInteger(), nullable=False),
        sa.Column('active_connections', sa.Integer(), nullable=False),
        sa.Column('total_connections', sa.BigInteger(), nullable=False),
        sa.PrimaryKeyConstraint('listener_id'),
        sa.ForeignKeyConstraint(['listener_id'], ['listener.id'],
                                name='fk_listener_statistics_listener_id')
    )

    op.create_table(
        'amphora',
        # id should come from the service providing the amphora (i.e. nova)
        sa.Column('id', sa.String(36), nullable=False, autoincrement=False),
        sa.Column('host_id', sa.String(36), nullable=False),
        sa.Column('status', sa.String(36), nullable=False),
        sa.PrimaryKeyConstraint('id'),
        sa.ForeignKeyConstraint(
            ['status'], ['provisioning_status.name'],
            name='fk_container_provisioning_status_name')
    )

    op.create_table(
        'load_balancer_amphora',
        sa.Column('amphora_id', sa.String(36), nullable=False),
        sa.Column('load_balancer_id', sa.String(36), nullable=False),
        sa.ForeignKeyConstraint(
            ['load_balancer_id'], ['load_balancer.id'],
            name='fk_load_balancer_amphora_load_balancer_id'),
        sa.ForeignKeyConstraint(['amphora_id'],
                                ['amphora.id'],
                                name='fk_load_balancer_amphora_id'),
        sa.PrimaryKeyConstraint('amphora_id', 'load_balancer_id')
    )



# Copyright 2014 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from unittest import mock

from oslo_config import cfg
from oslo_config import fixture as oslo_fixture

import octavia.certificates.common.auth.barbican_acl as barbican_acl
import octavia.certificates.manager.barbican as barbican_cert_mgr
from octavia.common import keystone
import octavia.tests.unit.base as base

CONF = cfg.CONF


class TestBarbicanACLAuth(base.TestCase):

    def setUp(self):
        super().setUp()
        # Reset the client
        keystone._SESSION = None
        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        self.region_name = 'RegionOne'
        self.endpoint_type = 'publicURL'
        self.endpoint = 'barbican'
        self.conf.config(group="certificates", region_name=self.region_name)
        self.conf.config(group="certificates",
                         endpoint_type=self.endpoint_type)
        self.conf.config(group="certificates", endpoint=self.endpoint)

    @mock.patch('barbicanclient.client.Client')
    @mock.patch('keystoneauth1.session.Session')
    def test_get_barbican_client(self, mock_ksession, mock_client):
        session_mock = mock.Mock()
        mock_ksession.return_value = session_mock
        mock_client.return_value = mock.MagicMock()

        # Mock out the keystone session and get the client
        acl_auth_object = barbican_acl.BarbicanACLAuth()
        bc1 = acl_auth_object.get_barbican_client()

        mock_client.assert_called_once_with(session=session_mock,
                                            region_name=self.region_name,
                                            interface=self.endpoint_type)

        mock_client.reset_mock()
        # Getting the session again with new class should get the same object
        acl_auth_object2 = barbican_acl.BarbicanACLAuth()
        bc2 = acl_auth_object2.get_barbican_client()
        self.assertIs(bc1, bc2)

        mock_client.assert_not_called()

    def test_load_auth_driver(self):
        bcm = barbican_cert_mgr.BarbicanCertManager()
        self.assertIsInstance(bcm.auth, barbican_acl.BarbicanACLAuth)

    @mock.patch('barbicanclient.client.Client')
    @mock.patch('octavia.common.keystone.KeystoneSession')
    def test_ensure_secret_access(self, mock_ksession, mock_client):
        service_user_id = 'uuid1'
        client_mock = mock.MagicMock()
        mock_client.return_value = client_mock
        mock_ksession().get_service_user_id.return_value = service_user_id

        mock_acl = mock.MagicMock()
        client_mock.acls.get.return_value = mock_acl

        mock_read = mock.MagicMock()
        mock_read.users = []
        mock_acl.get.return_value = mock_read

        acl_auth_object = barbican_acl.BarbicanACLAuth()
        acl_auth_object.ensure_secret_access(mock.Mock(), mock.Mock())
        mock_acl.submit.assert_called_once()
        self.assertEqual([service_user_id], mock_read.users)

    @mock.patch('barbicanclient.client.Client')
    @mock.patch('octavia.common.keystone.KeystoneSession')
    def test_revoke_secret_access(self, mock_ksession, mock_client):
        service_user_id = 'uuid1'

        client_mock = mock.MagicMock()
        mock_client.return_value = client_mock
        mock_ksession().get_service_user_id.return_value = service_user_id

        mock_acl = mock.MagicMock()
        client_mock.acls.get.return_value = mock_acl

        mock_read = mock.MagicMock()
        mock_read.users = [service_user_id]
        mock_acl.get.return_value = mock_read

        acl_auth_object = barbican_acl.BarbicanACLAuth()
        acl_auth_object.revoke_secret_access(mock.Mock(), mock.Mock())
        mock_acl.submit.assert_called_once()

    @mock.patch('octavia.common.keystone.KeystoneSession')
    @mock.patch('barbicanclient.client.Client')
    @mock.patch('keystoneauth1.session.Session')
    def test_get_barbican_client_user_auth(self, mock_ksession, mock_client,
                                           mock_keystone):
        session_mock = mock.MagicMock()
        mock_ksession.return_value = session_mock
        acl_auth_object = barbican_acl.BarbicanACLAuth()
        acl_auth_object.get_barbican_client_user_auth(mock.Mock())

        mock_client.assert_called_once_with(session=session_mock,
                                            endpoint=self.endpoint)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Add member vnic_type

Revision ID: 8db7a6443785
Revises: 3097e55493ae
Create Date: 2024-03-29 20:34:37.263847

"""

from alembic import op
import sqlalchemy as sa

from octavia.common import constants

# revision identifiers, used by Alembic.
revision = '8db7a6443785'
down_revision = '3097e55493ae'


def upgrade():
    op.add_column(
        u'member',
        sa.Column(u'vnic_type', sa.String(64), nullable=False,
                  server_default=constants.VNIC_TYPE_NORMAL)
    )
